#
# data_source.py -- utilities to assist in making time-series plots
#
import numbers
from collections import deque

import numpy as np


class XYDataSource:
    """Monitor a data source efficiently.

    Uses a static array circular queue and a sliding window algorithm to
    efficiently keep track of the current set of points and it's minmax.
    """

    def __init__(self, arr_src, points=[], overwrite=False,
                 none_for_empty=False):
        """Constructor for Data Source

        Parameters
        ----------
        arr_src : ndarray
            (x, y) points to initialize the data source with

        points : array, list, tuple or sequence-like (optional)
            An array of points to initialize the data source with

        overwrite : bool (optional)
            If ``True`` the buffer will eventually overwrite old entries

        none_for_empty : bool (optional)
            If ``True`` then don't raise a ValueError for an empty buffer

        """
        self.buf = arr_src
        self.length = len(arr_src)
        self.overwrite = overwrite
        self.none_for_empty = none_for_empty

        # for circular queue
        self.front = self.length - 1
        self.rear = None

        self.limits = np.array([[0.0, 0.0], [0.0, 0.0]])

        self.set_points(points)

    def get_limits(self):
        """Return the limits of this current data source.  The return value
        looks like: [(xmin, ymin), (xmax, ymax)]
        """
        return np.copy(self.limits)

    def update_limits(self):
        """Mostly internal routine to update the limits.  Shouldn't need to
        be called explicitly in most cases.
        """
        if len(self) == 0:
            self.limits = np.array([[0.0, 0.0], [0.0, 0.0]])
        else:
            x_min, x_max = self.buf[self.rear][0], self.buf[self.front][0]
            y_min, y_max = self.slmm.get_minmax()
            self.limits = np.array([[x_min, y_min], [x_max, y_max]])

    def set_points(self, points):
        """Initialize the data source with a series of points.
        `points` should be a list/array/sequence of (x, y)
        """
        self.slmm = SlidingWindowMinMax()
        self.front = self.length - 1
        self.rear = None

        self.add_points(points)

    def add_points(self, points):
        """Add a series of points.
        `points` should be a list/array/sequence of (x, y)
        """
        for pt in points:
            self._add(pt)
        self.update_limits()

    def is_fullp(self):
        """Returns True if there is no more room in the buffer."""
        front = (self.front + 1) % self.length
        return front == self.rear

    def _add(self, pt):
        x, y = pt
        # skip bogus values
        if not isinstance(y, numbers.Number) or not np.isfinite(y):
            return

        front = (self.front + 1) % self.length

        if front == self.rear:
            if not self.overwrite:
                raise ValueError("Buffer is full")
            # circular queue full, need to expunge an old element
            _x, _y = self.buf[self.rear]
            self.slmm.remove_head(_y)
            self.rear = (self.rear + 1) % self.length

        self.front = front
        if self.rear is None:
            self.rear = self.front

        self.buf[self.front, :] = pt
        self.slmm.add_tail(y)

    def add(self, pt, update_limits=True):
        """Add a single data point and update the plot.
        If the number of points exceeds the `length` of the data source,
        the oldest point will be ejected.
        If `update_limits` is `True` (the default) then the limits of the
        current data set are updated.
        """
        self._add(pt)
        if update_limits:
            self.update_limits()

    append = add
    push = add

    def peek(self):
        """Get the latest data point.  Will return `None` if there are no
        points recorded.
        """
        if len(self) == 0:
            if self.none_for_empty:
                return None
            raise ValueError("Buffer is empty")
        return self.buf[self.front]

    get_latest = peek

    def pop(self):
        """Get and remove the latest data point.
        Will return `None` if there are no points recorded.
        """
        if len(self) == 0:
            if self.none_for_empty:
                return None
            raise ValueError("Buffer is empty")
        pt = self.buf[self.front]
        if self.rear == self.front:
            self.rear = None
        else:
            self.front = self.length - 1 if self.front == 0 else self.front - 1
        return pt

    def peek_rear(self):
        """Get the earliest data point.  Will return `None` if there are no
        points recorded.
        """
        if len(self) == 0:
            if self.none_for_empty:
                return None
            raise ValueError("Buffer is empty")
        return self.buf[self.rear]

    def pop_rear(self):
        """Get and remove the earliest data point.
        Will return `None` if there are no points recorded.
        """
        if len(self) == 0:
            if self.none_for_empty:
                return None
            raise ValueError("Buffer is empty")
        pt = self.buf[self.rear]
        if self.rear == self.front:
            self.rear = None
        else:
            self.rear = (self.rear + 1) % self.length
        return pt

    def get_points(self):
        """Get the entire set of data points as an `ndarray`.
        """
        n = len(self)
        arr = np.zeros((n, 2), dtype=float)
        if n == 0:
            return arr
        if self.front >= self.rear:
            arr[:n, :] = self.buf[self.rear:self.front + 1, :]
        else:
            m = self.length - self.rear
            arr[0:m, :] = self.buf[self.rear:self.length, :]
            arr[m:n, :] = self.buf[0:self.front + 1, :]
        return arr

    @property
    def points(self):
        return self.get_points()

    def __len__(self):
        if self.rear is None:
            return 0
        if self.front >= self.rear:
            return self.front - self.rear + 1
        else:
            return self.length - self.rear + self.front + 1


class SlidingWindowMinMax:
    """Class to efficiently keep track of the minmax values of a changing
    data set.
    """

    def __init__(self):
        self.min_deque = deque()
        self.max_deque = deque()

    def get_minmax(self):
        return (self.min_deque[0], self.max_deque[0])

    def add_tail(self, val):
        while len(self.min_deque) > 0 and val < self.min_deque[-1]:
            self.min_deque.pop()
        self.min_deque.append(val)

        while len(self.max_deque) > 0 and val > self.max_deque[-1]:
            self.max_deque.pop()
        self.max_deque.append(val)

    def remove_head(self, val):
        if val < self.min_deque[0]:
            raise ValueError(f"Wrong value (min: {self.min_deque[0]}, val: {val})")
        elif val == self.min_deque[0]:
            self.min_deque.popleft()

        if val > self.max_deque[0]:
            raise ValueError(f"Wrong value (max: {self.max_deque[0]}, val: {val})")
        elif val == self.max_deque[0]:
            self.max_deque.popleft()


def update_plot_from_source(dsrc, xyplot, update_limits=False):
    """Update the associated plot with the current set of points.
    If `update_limits` is `True` then the plot limits will be updated
    with the current limits of the data.
    """
    arr = dsrc.get_points()
    if update_limits:
        limits = dsrc.get_limits()
        xyplot.plot(arr, limits=limits)
    else:
        xyplot.plot(arr)



#
# Channel.py -- Channel class for the Ginga reference viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import time

from ginga.misc import Bunch, Datasrc, Callback, Future, Settings
from ginga.util import viewer as gviewer


class ChannelError(Exception):
    pass


class Channel(Callback.Callbacks):
    """Class to manage a channel.

    Parameters
    ----------
    name : str
        Name of the channel.

    fv : `~ginga.rv.Control.GingaShell`
        The reference viewer shell.  This is the central control object
        for the Ginga Reference Viewer.

    settings : `~ginga.misc.Settings.SettingGroup`
        Channel settings.

    datasrc : `~ginga.misc.Datasrc.Datasrc`
        Data cache.

    """
    def __init__(self, name, fv, settings, datasrc=None):
        super(Channel, self).__init__()

        self.logger = fv.logger
        self.fv = fv
        self.settings = settings
        self.logger = fv.logger

        # CHANNEL ATTRIBUTES
        self.name = name
        self.widget = None
        self.container = None
        self.workspace = None
        self.opmon = None
        # this is the image viewer we are connected to
        self.fitsimage = None
        # this is the currently active viewer
        self.viewer = None
        self.viewers = []
        self.viewer_dict = {}
        if datasrc is None:
            num_images = self.settings.get('numImages', 1)
            datasrc = Datasrc.Datasrc(num_images)
        self.datasrc = datasrc
        self.cursor = -1
        self.history = []
        self.image_index = {}
        # external entities can attach stuff via this attribute
        self.extdata = Bunch.Bunch()

        self._configure_sort()
        self.settings.get_setting('sort_order').add_callback(
            'set', self._sort_changed_ext_cb)

    def connect_viewer(self, viewer):
        """Add a viewer to the set of viewers for this channel."""
        if viewer not in self.viewers:
            self.viewers.append(viewer)
            self.viewer_dict[viewer.vname] = viewer

    def move_image_to(self, imname, channel):
        """Move a data object named `imname` from this channel to another
        channel.

        Parameters
        ----------
        imname : str
            Name of the data in this channel (names are unique per-channel)

        channel : `~ginga.rv.Channel.Channel`
            Channel to which we will move the data

        Callbacks
        ---------
        Will invoke `add-image` in the receiving channel if a data object
        with that name does not exist there already (see add_image_update()).

        Will invoke `remove-image` in this channel (see remove_image()).
        """
        if self == channel:
            return

        self.copy_image_to(imname, channel)
        self.remove_image(imname)

    def copy_image_to(self, imname, channel, silent=False):
        """Copy a data object named `imname` from this channel to another
        channel.

        Parameters
        ----------
        imname : str
            Name of the data in this channel (names are unique per-channel)

        channel : `~ginga.rv.Channel.Channel`
            Channel to which we will copy the data

        Callbacks
        ---------
        Will invoke `add-image` in the receiving channel if a data object
        with that name does not exist there already (see add_image_update()).
        """
        if self == channel:
            return

        if imname in channel:
            # image with that name is already there
            return

        # transfer image info
        info = self.image_index[imname]
        was_not_there_already = channel._add_info(info)

        try:
            image = self.datasrc[imname]

        except KeyError:
            return

        if was_not_there_already:
            channel.datasrc[imname] = image

            if not silent:
                self.fv.gui_do(channel.add_image_update, image, info,
                               update_viewer=False)

    def remove_image(self, imname):
        """Remove a data object named `imname` from this channel.

        Parameters
        ----------
        imname : str
            Name of the data in this channel (names are unique per-channel)

        Callbacks
        ---------
        Will invoke `remove-image` including the channel name, data object
        name and path.
        """
        info = self.image_index[imname]
        self.remove_history(imname)

        if imname in self.datasrc:
            image = self.datasrc[imname]
            self.datasrc.remove(imname)

            # update viewer if we are removing the currently displayed image
            cur_image = self.viewer.get_dataobj()
            if cur_image == image:
                self.refresh_cursor_image()

        self.fv.make_async_gui_callback('remove-image', self.name,
                                        info.name, info.path)

        return info

    def get_image_names(self):
        """Return the list of data items in this channel."""
        return [info.name for info in self.history]

    def get_loaded_image(self, imname):
        """Get an data object from this channel's memory cache.

        Parameters
        ----------
        imname : str
            Key, usually image name and extension.

        Returns
        -------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object.

        Raises
        ------
        KeyError
            If the named data item is not in the memory cache.

        """
        image = self.datasrc[imname]
        return image

    def add_image(self, image, silent=False, bulk_add=False):
        """Add a data object to this channel.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object to be added.

        silent : bool (optional, defaults to `False`)
            Indicates a "silent add", in which case the callback is
            suppressed.

        bulk_add : bool (optional, defaults to `False`)
            Indicates a "bulk add", in which the callback is not
            suppressed, but the channel viewer will not be updated.

        Callbacks
        ---------
        Will invoke `add-image`, if `silent` is `False`
        (see add_image_update()).
        """
        imname = image.get('name', None)
        if imname is None:
            raise ValueError("image has no name")

        self.logger.debug("Adding image '%s' in channel %s" % (
            imname, self.name))

        self.datasrc[imname] = image

        # Has this image been loaded into a channel before?
        info = image.get('image_info', None)
        if info is None:
            # No
            idx = image.get('idx', None)
            path = image.get('path', None)
            nothumb = image.get('nothumb', False)
            image_loader = image.get('image_loader', None)
            image_future = image.get('image_future', None)
            info = self.add_history(imname, path,
                                    image_loader=image_loader,
                                    image_future=image_future,
                                    idx=idx, nothumb=nothumb)
            image.set(image_info=info)

        # add an image profile if one is missing
        profile = self.get_image_profile(image)
        info.profile = profile

        if not silent:
            self.add_image_update(image, info,
                                  update_viewer=not bulk_add)

    def add_image_info(self, info):
        """Add a data object's metadata to this channel.

        This function is used to add enough metadata about a data
        object to the channel so that the object can be later loaded
        and viewed from the channel on demand.  To do so, the bunch
        passed as `info` should include as a minimum:

        * name: name of object
        * path: file path to object

        Optionally, an image loader or image future can be specified.
        If ommitted, the default loader is used and a future is created
        using that loader.

        Parameters
        ----------
        info : `~ginga.Bunch.Bunch`
            Metadata about a data object to be added.

        Callbacks
        ---------
        Will invoke `add-image-info` (see add_history()).
        """
        image_loader = info.get('image_loader', self.fv.load_image)

        # create an image_future if one does not exist
        image_future = info.get('image_future', None)
        if (image_future is None) and (info.path is not None):
            image_future = Future.Future()
            image_future.freeze(image_loader, info.path)

        self.add_history(info.name, info.path,
                         image_loader=image_loader,
                         image_future=image_future)

    def get_image_info(self, imname):
        """Return metadata about a data object in this channel.

        Parameters
        ----------
        imname : str
            Name of the data in this channel (names are unique per-channel)

        Returns
        -------
        info : `~ginga.Bunch.Bunch`
            Metadata about a data object to be added.
        """
        return self.image_index[imname]

    def update_image_info(self, image, info):
        """Update metadata about a data object in this channel.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object to be updated.

        info : `~ginga.Bunch.Bunch`
            Metadata about the data object to be updated.

        Callbacks
        ---------
        Will invoke `add-image-info` including the updated metadata.
        """
        imname = image.get('name', None)
        if (imname is None) or (imname not in self.image_index):
            return False

        # don't update based on image name alone--actual image must match
        try:
            my_img = self.get_loaded_image(imname)
            if my_img is not image:
                return False

        except KeyError:
            return False

        # update the info record
        iminfo = self.get_image_info(imname)
        iminfo.update(info)

        self.fv.make_async_gui_callback('add-image-info', self, iminfo)
        return True

    def add_image_update(self, image, info, update_viewer=False):
        """Update metadata about a data object in this channel.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object to be updated.

        info : `~ginga.Bunch.Bunch`
            Metadata about the data object to be updated.

        update_viewer : bool (optional, defaults to `False`)
            If `True`, will also update the channel viewer if the data
            object is the same as the most recent object in the channel.

        Callbacks
        ---------
        Will invoke `add-image` including the data name, object and metadata.
        """
        self.fv.make_async_gui_callback('add-image', self.name, image, info)

        if not update_viewer:
            return

        current = self.datasrc.youngest()
        curname = current.get('name')
        self.logger.debug("image=%s youngest=%s" % (image.get('name'), curname))
        if current != image:
            return

        # switch to current image?
        if self.settings['switchnew']:
            self.logger.debug("switching to new image '%s'" % (curname))
            self.switch_image(image)

        if self.settings['raisenew']:
            channel = self.fv.get_current_channel()
            if channel != self:
                self.fv.change_channel(self.name)

    def refresh_cursor_image(self):
        """Refresh the channel viewer to the current item pointed to
        by the channel cursor.

        Mostly internal routine used when the channel cursor is changed.
        """
        if self.cursor < 0:
            self.viewer.clear()
            self.fv.channel_image_updated(self, None)
            return

        info = self.history[self.cursor]
        if info.name in self.datasrc:
            # object still in memory
            data_obj = self.datasrc[info.name]
            self.switch_image(data_obj)

        else:
            self.switch_name(info.name)

    def prev_image(self, loop=True):
        """Move the channel cursor to the previous data object in this
        channel.

        Parameters
        ----------
        loop : bool (optional, defaults to `True`)
            If `True`, will loop around to the end of the channel's
            data objects if the cursor is at the start.
        """
        self.logger.debug("Previous image")
        if self.cursor <= 0:
            n = len(self.history) - 1
            if (not loop) or (n < 0):
                self.logger.error("No previous image!")
                return True
            self.cursor = n
        else:
            self.cursor -= 1

        self.refresh_cursor_image()
        return True

    def next_image(self, loop=True):
        """Move the channel cursor to the next data object in this
        channel.

        Parameters
        ----------
        loop : bool (optional, defaults to `True`)
            If `True`, will loop around to the start of the channel's
            data objects if the cursor is at the end.
        """
        self.logger.debug("Next image")
        n = len(self.history) - 1
        if self.cursor >= n:
            if (not loop) or (n < 0):
                self.logger.error("No next image!")
                return True
            self.cursor = 0
        else:
            self.cursor += 1

        self.refresh_cursor_image()
        return True

    def _add_info(self, info):
        """Internal method used to add metadata to the channel's index
        of data objects.
        """
        if info.name in self.image_index:
            # image info is already present
            return False

        self.history.append(info)
        self.image_index[info.name] = info

        if self.hist_sort is not None:
            self.history.sort(key=self.hist_sort)

        self.fv.make_async_gui_callback('add-image-info', self, info)

        # image was newly added
        return True

    def add_history(self, imname, path, idx=None,
                    image_loader=None, image_future=None,
                    nothumb=False):
        """Add metadata about a data object to this channel.

        See add_image_info() for use and additional information.

        Parameters
        ----------
        imname : str
            Unique name (to this channel) of a data object to be added.

        path : str
            Path to the data object in storage.

        idx : int or str (optional, defaults to None)
            An optional index into the file, indicating an HDU or logical
            subunit of the file to load.

        image_loader : function (optional, defaults to None)
            An optional loader for this data item.  If `None` is passed,
            then defaults to the default loader.

        image_future : `~ginga.misc.Future.Future` (optional, default: None)
            An optional image future used to reload the image if it is
            unloaded from memory.  If `None` is passed, then defaults to
            a future using the default loader.

        nothumb : bool (optional, default: False)
            True if no thumbnail should be generated for this object.

        Returns
        -------
        info : `~ginga.Bunch.Bunch`
            A record of the metadata about the data object.

        Callbacks
        ---------
        Will invoke `add-image-info` including the updated metadata.
        """

        if imname not in self.image_index:

            if image_loader is None:
                image_loader = self.fv.load_image
            # create an image_future if one does not exist
            if (image_future is None) and (path is not None):
                image_future = Future.Future()
                image_future.freeze(image_loader, path)

            info = Bunch.Bunch(name=imname, path=path,
                               idx=idx,
                               image_loader=image_loader,
                               image_future=image_future,
                               time_added=time.time(),
                               time_modified=None,
                               last_viewer_info=None,
                               profile=None,
                               nothumb=nothumb)
            self._add_info(info)

        else:
            # already in history
            info = self.image_index[imname]

            # refresh info, in case anything changed
            self.fv.make_async_gui_callback('add-image-info', self, info)

        return info

    def remove_history(self, imname):
        """Remove metadata about a data object in this channel.

        Parameters
        ----------
        imname : str
            Unique name (to this channel) of a data object.

        Callbacks
        ---------
        Will invoke `remove-image-info` including the metadata record.
        """
        if imname in self.image_index:
            info = self.image_index[imname]
            del self.image_index[imname]
            i = self.history.index(info)
            self.history.remove(info)

            # adjust cursor as necessary
            if i < self.cursor:
                self.cursor -= 1
            if self.cursor >= len(self.history):
                # loop
                self.cursor = min(0, len(self.history) - 1)

            self.fv.make_async_gui_callback('remove-image-info', self, info)

    def get_current_image(self):
        """Return the data object being viewed in this channel.

        Returns
        -------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object.
        """
        return self.viewer.get_dataobj()

    def view_object(self, dataobj):
        """View the data object (`dataobj`) in an appropriate channel
        viewer.

        This is a mostly internal method used to view a data object in
        the channel. See switch_image().

        Parameters
        ----------
        dataobj : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object to be viewed in an appropriate channel viewer.
        """
        # see if a viewer has been used on this object before
        vinfo = None
        obj_name = dataobj.get('name')
        if obj_name in self.image_index:
            info = self.image_index[obj_name]
            vinfo = info.last_viewer_info

        if vinfo is not None:
            # use the viewer we used before
            viewers = [vinfo]
        else:
            # find available viewers that can view this kind of object
            viewers = gviewer.get_priority_viewers(dataobj)
            if len(viewers) == 0:
                raise ValueError("No viewers for this data object!")
            self.logger.debug("{} available viewers for this model".format(len(viewers)))

        # if there is only one viewer available, use it otherwise
        # pop-up a dialog and ask the user
        if len(viewers) == 1:
            self.open_with_viewer(viewers[0], dataobj)
            return

        msg = ("Multiple viewers are available for this data object. "
               "Please select one.")
        self.fv.gui_choose_viewer(msg, viewers, self.open_with_viewer,
                                  dataobj)

    def get_viewer(self, vname):
        """Return the channel viewer with the specified name.

        Parameters
        ----------
        vname : str
            Name of the type of viewer.

        Returns
        -------
        viewer : a ginga channel viewer of the specified type

        """
        # if we don't have this viewer type then install one in the channel
        if vname not in self.viewer_dict:
            vinfo = gviewer.get_vinfo(vname)
            self.fv.make_viewer(vinfo, self)

        viewer = self.viewer_dict[vname]
        return viewer

    def open_with_viewer(self, vinfo, dataobj):
        # if we don't have this viewer type then install one in the channel
        if vinfo.name not in self.viewer_dict:
            self.fv.make_viewer(vinfo, self)

        old_viewer, self.viewer = self.viewer, self.viewer_dict[vinfo.name]
        # find this viewer and raise it
        idx = self.viewers.index(self.viewer)
        self.widget.set_index(idx)

        self.fv.make_async_gui_callback('viewer-select', self,
                                        old_viewer, self.viewer)

        # and load the data
        if self.viewer.get_dataobj() is not dataobj:
            self.viewer.set_dataobj(dataobj)

        obj_name = dataobj.get('name')
        if obj_name in self.image_index:
            info = self.image_index[obj_name]
            # record viewer last used to view this object
            info.last_viewer_info = vinfo
            if info in self.history:
                # update cursor to match dataobj
                self.cursor = self.history.index(info)

        self.fv.channel_image_updated(self, dataobj)

    def switch_image(self, image):
        """Switch to data object (`image`) in this channel.

        Parameters
        ----------
        image : subclass of `~ginga.BaseImage.ViewerObjectBase`
            Data object in this channel to be viewed in an appropriate
            channel viewer.
        """
        curimage = self.get_current_image()
        if curimage == image:
            self.logger.debug("Apparently no need to set channel viewer.")
            return

        self.logger.debug("updating viewer...")
        self.view_object(image)

    def switch_name(self, imname):
        """Switch to data object named by `imname` in this channel.

        Parameters
        ----------
        imname : str
            Unique name of a data object in this channel.
        """
        if imname in self.datasrc:
            # Image is still in the heap
            image = self.datasrc[imname]
            self.switch_image(image)
            return

        if not (imname in self.image_index):
            errmsg = "No image by the name '%s' found" % (imname)
            self.logger.error("Can't switch to image '%s': %s" % (
                imname, errmsg))
            raise ChannelError(errmsg)

        # Do we have a way to reconstruct this image from a future?
        info = self.image_index[imname]
        if info.image_future is not None:
            self.logger.info("Image '%s' is no longer in memory; attempting "
                             "image future" % (imname))

            # TODO: recode this--it's a bit messy
            def _switch(image):
                # this will be executed in the gui thread
                self.add_image(image, silent=True)
                self.switch_image(image)

                # reset modified timestamp
                info.time_modified = None
                self.fv.make_async_gui_callback('add-image-info', self, info)

            def _load_n_switch(imname, path, image_future):
                # this will be executed in a non-gui thread
                # reconstitute the image
                image = self.fv.error_wrap(image_future.thaw)
                if isinstance(image, Exception):
                    errmsg = "Error reconstituting image: %s" % (str(image))
                    self.logger.error(errmsg)
                    raise image

                profile = info.get('profile', None)
                if profile is None:
                    profile = self.get_image_profile(image)
                    info.profile = profile
                # perpetuate some of the image metadata
                image.set(image_future=image_future, name=imname, path=path,
                          image_info=info, profile=profile)

                self.fv.gui_do(_switch, image)

            self.fv.nongui_do(_load_n_switch, imname, info.path,
                              info.image_future)

        elif info.path is not None:
            # Do we have a path? We can try to reload it
            self.logger.debug("Image '%s' is no longer in memory; attempting "
                              "to load from %s" % (imname, info.path))

            #self.fv.load_file(path, chname=chname)
            self.fv.nongui_do(self.load_file, info.path, chname=self.name)

        else:
            raise ChannelError("No way to recreate image '%s'" % (imname))

    def _configure_sort(self):
        self.hist_sort = lambda info: info.time_added
        # set sorting function
        sort_order = self.settings.get('sort_order', 'loadtime')
        if sort_order == 'alpha':
            # sort history alphabetically
            self.hist_sort = lambda info: info.name

    def _sort_changed_ext_cb(self, setting, value):
        self._configure_sort()

        self.history.sort(key=self.hist_sort)

    def get_image_profile(self, image):
        """Get the image profile for data object `image`.

        The image profile is not an ICC profile, but rather a set of
        settings that the image was last viewed with.  The settings can
        be saved and restored when an image is viewed according to the
        channel preferences under Reset (Viewer) and Rememeber (Image).
        """
        profile = image.get('profile', None)
        if profile is None:
            profile = Settings.SettingGroup()
            image.set(profile=profile)
        return profile

    def __len__(self):
        """Return the number of the items added to this channel."""
        return len(self.history)

    def __contains__(self, imname):
        """Returns `True` if data object named `imname` is in this channel."""
        return imname in self.image_index

    def __getitem__(self, imname):
        """Returns metadata about the data object named `imname`."""
        return self.image_index[imname]

# END



#! /usr/bin/env python3
"""
This example shows the use of the Ginga matplotlib-based plot viewer.
"""
import sys
from argparse import ArgumentParser

import numpy as np

import ginga.toolkit as ginga_toolkit
from ginga.misc import log
from ginga.plot.Plotable import Plotable

# window size
dims = (600, 440)


def main(options, args):
    logger = log.get_logger("sine", options=options)

    if options.toolkit is None:
        logger.error("Please choose a GUI toolkit with -t option")

    # decide our toolkit, then import
    ginga_toolkit.use(options.toolkit)

    # now we can import
    from ginga.gw import Widgets
    from ginga.plot import PlotView

    app = Widgets.Application(logger=logger)

    def quit(*args):
        logger.info("Top window closed.")
        sys.exit()

    app.add_callback('shutdown', quit)
    if hasattr(Widgets, 'Page'):
        page = Widgets.Page("Plot sine2")
        app.add_window(page)
        top = Widgets.TopLevel("Plot sine2")
        page.add_dialog(top)
    else:
        top = Widgets.TopLevel("Plot sine2")
        app.add_window(top)
    top.add_callback('close', quit)

    vbox = Widgets.VBox()
    vbox.set_spacing(1)
    vbox.set_border_width(2)
    win_wd, win_ht = dims[:2]
    viewer = PlotView.CanvasView(logger=logger)
    viewer.set_enter_focus(True)

    bd = viewer.get_bindings()
    bd.enable_all(True)

    settings = viewer.get_settings()
    settings.set(plot_show_mode=True)

    viewer.set_limits([(-2, -1), (2, 1)])

    plot_w = viewer.get_ginga_widget()
    vbox.add_widget(plot_w, stretch=1)

    hbox = Widgets.HBox()
    btn = Widgets.Button("Quit")
    btn.add_callback('activated', quit)
    hbox.add_widget(Widgets.Label(''), stretch=1)
    hbox.add_widget(btn, stretch=0)
    vbox.add_widget(hbox, stretch=0)

    top.set_widget(vbox)

    x_data = np.arange(-3, 3.5, 0.01)
    y_sin = np.sin(x_data)
    y_cos = np.cos(x_data)

    plotable = Plotable(logger=logger)
    plotable.plot_line(np.array((x_data, y_sin)).T, color='red', tag='sin')
    plotable.plot_line(np.array((x_data, y_cos)).T, color='blue', tag='cos')
    plotable.set(title="Sin/Cos", x_axis_label="X", y_axis_label="Y",
                 grid=True)

    viewer.set_dataobj(plotable)
    # viewer.set_ranges(x_range=(-2.0, 3.5), y_range=(-1.0, 1.0))

    top.resize(win_wd, win_ht)
    top.show()
    top.raise_()

    try:
        app.mainloop()

    except KeyboardInterrupt:
        top.close()


if __name__ == "__main__":

    # Parse command line options
    argprs = ArgumentParser("Example of PlotView viewer")

    argprs.add_argument("-t", "--toolkit", dest="toolkit", metavar="NAME",
                        default='qt',
                        help="Choose GUI toolkit (gtk|qt)")
    argprs.add_argument("-r", "--renderer", dest="renderer", metavar="NAME",
                        default=None,
                        help="Choose renderer (pil|agg|opencv|cairo|qt)")
    log.addlogopts(argprs)

    (options, args) = argprs.parse_known_args(sys.argv[1:])

    main(options, args)



#
# plot2d.py -- mode for manipulating 2D plots in PlotView based viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.plot.PlotView import PlotViewBase
from ginga.misc import Bunch
from ginga.modes.mode_base import Mode


class Plot2DMode(Mode):
    """
    Plot2D Mode enables bindings that can set the pan position and zoom level
    (scale) in a Matplotlib plot viewer (ginga.plot.PlotView module).

    Enter the mode by
    -----------------
    * Space, then "p"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * Shift + left click : set pan position
    * middle click : set pan position
    * scroll : zoom in/out
    * ctrl + scroll : zoom in/out X axis only
    * shift + scroll : zoom in/out Y axis only (On MacOS, trackpad only)
    * alt + scroll(mouse) : zoom in/out Y axis only (Option key on Macs)
    * alt + scroll : zoom in/out at cursor

    Keystroke bindings
    ------------------

    Zooming
    -------
    * equals : zoom in one zoom level
    * ctrl + equals : zoom in X axis one zoom level
    * plus (shift + equals) : zoom in Y axis one zoom level
    * minus : zoom out one zoom level
    * ctrl + minus : zoom out X axis one zoom level
    * underscore (shift + minus): zoom out Y axis one zoom level
    * 9 : zoom out maintaining cursor position
    * ctrl + 9 : zoom out X axis maintaining cursor position
    * left paren (shift + 9): zoom out Y axis maintaining cursor position
    * 0 : zoom in maintaining cursor position
    * ctrl + 0 : zoom in X axis maintaining cursor position
    * right paren (shift + 0): zoom in Y axis maintaining cursor position
    * backquote : zoom X and Y axes to fit window
    * 1 : zoom X axis only to fit window
    * 2 : zoom Y axis only to fit window
    * k : set lower X range to X value at cursor
    * l : set upper X range to X value at cursor
    * K : set lower Y range to Y value at cursor
    * L : set upper Y range to Y value at cursor
    * g : toggle grid on/off
    * f : toggle legend on/off

    Panning
    -------
    * left arrow : pan left
    * right arrow : pan right
    * up arrow : pan up
    * down arrow : pan down

    """

    # Needs to be set by reference viewer (via set_shell_ref) before any
    # channel viewers are created
    fv = None

    @classmethod
    def set_shell_ref(cls, fv):
        cls.fv = fv

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, PlotViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_plot2d=['__p', None, 'plot2d'],

            ms_panset2d=['plot2d+middle', 'plot2d+shift+left'],

            sc_zoom2d=['plot2d+scroll'],
            sc_zoom2d_x=['plot2d+ctrl+scroll'],
            sc_zoom2d_y=['plot2d+shift+scroll'],
            sc_zoom2d_cursor=['plot2d+win+scroll'],

            pa_zoom=['plot2d+pan'],
            #pa_zoom_origin=['plot2d+shift+pan'])

            kp_pan_left=['plot2d+left'],
            kp_pan_right=['plot2d+right'],
            kp_pan_up=['plot2d+up'],
            kp_pan_down=['plot2d+down'],

            kp_zoom_in=['plot2d+='],
            kp_zoom_in_x=['plot2d+ctrl+='],
            kp_zoom_in_y=['plot2d+shift++'],
            kp_zoom_out=['plot2d+-'],
            kp_zoom_out_x=['plot2d+ctrl+-'],
            kp_zoom_out_y=['plot2d+shift+_'],

            kp_zoom_cursor_in=['plot2d+0'],
            kp_zoom_cursor_in_x=['plot2d+ctrl+0'],
            kp_zoom_cursor_in_y=['plot2d+shift+)'],
            kp_zoom_cursor_out=['plot2d+9'],
            kp_zoom_cursor_out_x=['plot2d+ctrl+9'],
            kp_zoom_cursor_out_y=['plot2d+shift+('],

            kp_zoom_fit=['plot2d+backquote'],
            kp_zoom_fit_x=['plot2d+1'],
            kp_zoom_fit_y=['plot2d+2'],
            kp_toggle_legend=['plot2d+f'],
            kp_toggle_grid=['plot2d+g'],

            kp_cut_x_lo=['plot2d+k'],
            kp_cut_x_hi=['plot2d+l'],
            kp_cut_y_lo=['plot2d+K'],
            kp_cut_y_hi=['plot2d+L'],

            plot_zoom_rate=1.2,
            plot_pan_pct=0.10,
        )

        # for interactive features
        self.can = Bunch.Bunch(zoom=True, pan=True)

    @property
    def canpan(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('pan')

    @property
    def canzoom(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('zoom')

    def __str__(self):
        return 'plot2d'

    def start(self):
        pass

    def stop(self):
        pass

    #####  SCROLL ACTION CALLBACKS #####

    def __zoom(self, viewer, delta_x, delta_y, origin=None):

        if origin is not None:
            cur_x, cur_y = origin
            if None not in origin:
                viewer.zoom_plot_at_cursor(cur_x, cur_y, delta_x, delta_y)
        else:
            viewer.zoom_plot(delta_x, delta_y)

        return True

    def __zoom2d(self, viewer, event, axis='xy', origin=None):
        if not self.canzoom:
            return False

        event.accept()
        rev = self.settings.get('zoom_scroll_reverse', False)
        direction = self.get_direction(event.direction, rev=rev)

        if direction == 'up':
            delta = self.settings['plot_zoom_rate'] ** -2

        elif direction == 'down':
            delta = self.settings['plot_zoom_rate'] ** 2

        delta_x = delta_y = delta
        if axis == 'x':
            # only horizontal
            delta_y = 1.0
        elif axis == 'y':
            # only vertical
            delta_x = 1.0

        return self.__zoom(viewer, delta_x, delta_y, origin=origin)

    def sc_zoom2d(self, viewer, event):
        """Can be set as the callback function for the 'scroll'
        event to zoom the plot.
        """
        return self.__zoom2d(viewer, event)

    def sc_zoom2d_x(self, viewer, event):
        return self.__zoom2d(viewer, event, axis='x')

    def sc_zoom2d_y(self, viewer, event):
        return self.__zoom2d(viewer, event, axis='y')

    def sc_zoom2d_cursor(self, viewer, event):
        return self.__zoom2d(viewer, event,
                             origin=(event.data_x, event.data_y))

    def sc_zoom2d_cursor_x(self, viewer, event):
        return self.__zoom2d(viewer, event, axis='x',
                             origin=(event.data_x, event.data_y))

    def sc_zoom2d_cursor_y(self, viewer, event):
        return self.__zoom2d(viewer, event, axis='y',
                             origin=(event.data_x, event.data_y))

    #####  MOUSE ACTION CALLBACKS #####

    def ms_panset2d(self, viewer, event, data_x, data_y):
        """An interactive way to set the pan position.  The location
        (data_x, data_y) will be centered in the window.

        Can be set as the callback function for the 'button-press'
        event to pan the plot with middle-click.
        """
        if not self.canpan:
            return False

        event.accept()
        if event.state != 'down':
            return

        if not self.can.pan:
            return

        if None not in [data_x, data_y]:
            viewer.set_pan(data_x, data_y)

        return True

    ##### GESTURE ACTION CALLBACKS #####

    def pa_zoom(self, viewer, event, msg=True):
        """Interactively zoom the image by a pan gesture.
        (the back end must support gestures)
        """
        print("pa zoom")
        if not self.canzoom:
            return False
        event.accept()
        print(event, dir(event))
        event = self._pa_synth_scroll_event(event)
        print(event)
        if event.state == 'move':
            #self._sc_zoom(viewer, event, msg=msg, origin=None)
            pass

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_zoom_out(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta = self.settings['plot_zoom_rate'] ** 2
        delta_x = delta_y = delta
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_out_x(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = self.settings['plot_zoom_rate'] ** 2
        delta_y = 1.0
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_out_y(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = 1.0
        delta_y = self.settings['plot_zoom_rate'] ** 2
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_in(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta = self.settings['plot_zoom_rate'] ** -2
        delta_x = delta_y = delta
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_in_x(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = self.settings['plot_zoom_rate'] ** -2
        delta_y = 1.0
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_in_y(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = 1.0
        delta_y = self.settings['plot_zoom_rate'] ** -2
        return self.__zoom(viewer, delta_x, delta_y)

    def kp_zoom_cursor_out(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta = self.settings['plot_zoom_rate'] ** 2
        delta_x = delta_y = delta
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_cursor_out_x(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = self.settings['plot_zoom_rate'] ** 2
        delta_y = 1.0
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_cursor_out_y(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = 1.0
        delta_y = self.settings['plot_zoom_rate'] ** 2
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_cursor_in(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta = self.settings['plot_zoom_rate'] ** -2
        delta_x = delta_y = delta
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_cursor_in_x(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = self.settings['plot_zoom_rate'] ** -2
        delta_y = 1.0
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_cursor_in_y(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        delta_x = 1.0
        delta_y = self.settings['plot_zoom_rate'] ** -2
        return self.__zoom(viewer, delta_x, delta_y, origin=(data_x, data_y))

    def kp_zoom_fit(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        viewer.zoom_fit(axis='xy')

    def kp_zoom_fit_x(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        viewer.zoom_fit(axis='x')

    def kp_zoom_fit_y(self, viewer, event, data_x, data_y):
        if not self.canzoom:
            return False
        event.accept()
        viewer.zoom_fit(axis='y')

    def kp_pan_left(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        pct = self.settings['plot_pan_pct']
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        x_rng = abs(x_hi - x_lo)
        x_delta, y_delta = -(x_rng * pct), 0.0
        viewer.pan_delta(x_delta, y_delta)

    def kp_pan_right(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        pct = self.settings['plot_pan_pct']
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        x_rng = abs(x_hi - x_lo)
        x_delta, y_delta = (x_rng * pct), 0.0
        viewer.pan_delta(x_delta, y_delta)

    def kp_pan_up(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        pct = self.settings['plot_pan_pct']
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        y_rng = abs(y_hi - y_lo)
        x_delta, y_delta = 0.0, (y_rng * pct)
        viewer.pan_delta(x_delta, y_delta)

    def kp_pan_down(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        pct = self.settings['plot_pan_pct']
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        y_rng = abs(y_hi - y_lo)
        x_delta, y_delta = 0.0, -(y_rng * pct)
        viewer.pan_delta(x_delta, y_delta)

    def kp_cut_x_lo(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        viewer.set_ranges(x_range=(data_x, x_hi))

    def kp_cut_x_hi(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        viewer.set_ranges(x_range=(x_lo, data_x))

    def kp_cut_y_lo(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        viewer.set_ranges(y_range=(data_y, y_hi))

    def kp_cut_y_hi(self, viewer, event, data_x, data_y):
        if not self.canpan:
            return False
        event.accept()
        (x_lo, x_hi), (y_lo, y_hi) = viewer.get_ranges()
        viewer.set_ranges(y_range=(y_lo, data_y))

    def kp_toggle_grid(self, viewer, event, data_x, data_y):
        event.accept()
        viewer.toggle_grid()

    def kp_toggle_legend(self, viewer, event, data_x, data_y):
        event.accept()
        viewer.toggle_legend()



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.canvas.CanvasObject import get_canvas_types
from ginga import trcalc
from ginga.gw import Widgets

from .base import Stage, StageAction


class Rotate(Stage):
    """
    Rotate a 2D image.

    'Rotate' (`rot_deg`) gives the angle of rotation for the image.

    'Clip' (`clip`) tells whether to clip the image at its existing size,
    or change size as needed to contain the entire image.

    'Add alpha' (`add_alpha`) will add an alpha channel to keep the
    parts of the expanded rotated image invisible.
    """

    _stagename = 'rotate'

    def __init__(self):
        super().__init__()

        self.dc = get_canvas_types()
        self.cropcolor = 'limegreen'
        self._rot_deg = 0.0
        self._clip = True
        self._add_alpha = False
        self.rot_obj = None
        self.viewer = None

        canvas = self.dc.DrawingCanvas()
        canvas.enable_edit(True)
        #canvas.set_callback('edit-event', self.edit_cb)
        canvas.set_draw_mode('edit')
        self.canvas = canvas

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')

        fr = Widgets.Frame("Rotate")

        captions = (('Rotate:', 'label', 'Rotate', 'entryset'),
                    ('_sp1', 'spacer', 'hbox1', 'hbox')
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')

        b.rotate.set_tooltip("Rotate the image around the pan position")
        b.rotate.set_text(str(self._rot_deg))
        b.rotate.add_callback('activated', self.rotate_cb)

        b.clip = Widgets.CheckBox("Clip")
        b.clip.set_state(self._clip)
        b.clip.set_tooltip("Clip rotated image")
        b.clip.add_callback('activated', self.clip_cb)
        b.hbox1.add_widget(b.clip, stretch=0)

        b.add_alpha = Widgets.CheckBox("Add alpha")
        b.add_alpha.set_state(self._add_alpha)
        b.add_alpha.set_tooltip("Add alpha channel to rotated image")
        b.add_alpha.add_callback('activated', self.add_alpha_cb)
        b.hbox1.add_widget(b.add_alpha, stretch=0)

        b.copy_from_viewer = Widgets.Button("Copy from viewer")
        b.copy_from_viewer.set_tooltip("Copy rotation setting from viewer")
        b.copy_from_viewer.add_callback('activated', self.copy_from_viewer_cb)
        b.hbox1.add_widget(b.copy_from_viewer, stretch=0)
        b.hbox1.add_widget(Widgets.Label(''), stretch=1)

        self.w.update(b)
        fr.set_widget(w)

        container.set_widget(fr)

        self.viewer = self.pipeline.get('viewer')
        self.canvas.set_surface(self.viewer)
        self.canvas.register_for_cursor_drawing(self.viewer)

        self.rot_obj = self.dc.Crosshair(100, 100)

    @property
    def rot_deg(self):
        return self._rot_deg

    @rot_deg.setter
    def rot_deg(self, val):
        self._rot_deg = val
        if self.gui_up:
            self.w.rotate.set_text(str(self._rot_deg))

    def _set_rotation(self, rot_deg):
        old_rot_deg, self.rot_deg = self._rot_deg, rot_deg
        self.pipeline.push(StageAction(self,
                                       dict(rot_deg=old_rot_deg),
                                       dict(rot_deg=self._rot_deg),
                                       descr="rotate angle"))
        self.pipeline.run_from(self)

    @property
    def clip(self):
        return self._clip

    @clip.setter
    def clip(self, tf):
        self._clip = tf
        if self.gui_up:
            self.w.clip.set_state(tf)

    def _set_clip(self, tf):
        old_clip, self.clip = self._clip, tf
        self.pipeline.push(StageAction(self,
                                       dict(clip=old_clip),
                                       dict(clip=self._clip),
                                       descr="change rotation clip"))
        self.pipeline.run_from(self)

    @property
    def add_alpha(self):
        return self._add_alpha

    @add_alpha.setter
    def add_alpha(self, tf):
        self._add_alpha = tf
        if self.gui_up:
            self.w.add_alpha.set_state(tf)

    def _set_add_alpha(self, tf):
        old_add_alpha, self._add_alpha = self._add_alpha, tf
        self.pipeline.push(StageAction(self,
                                       dict(add_alpha=old_add_alpha),
                                       dict(add_alpha=self._add_alpha),
                                       descr="change rotation alpha"))
        self.pipeline.run_from(self)

    def _get_state(self):
        return dict(rot_deg=self._rot_deg, clip=self._clip,
                    add_alpha=self._add_alpha)

    def rotate_cb(self, widget):
        rot_deg = float(widget.get_text().strip())
        self._set_rotation(rot_deg)

    def clip_cb(self, widget, tf):
        self._set_clip(tf)

    def add_alpha_cb(self, widget, tf):
        self._set_add_alpha(tf)

    def copy_from_viewer_cb(self, widget):
        rot_deg = self.viewer.get_rotation()
        self._set_rotation(rot_deg)

    def resume(self):
        # insert canvas, if not already
        p_canvas = self.viewer.get_canvas()
        if not p_canvas.has_object(self.canvas):
            p_canvas.add(self.canvas)

        if not self.canvas.has_object(self.rot_obj):
            self.canvas.add(self.rot_obj)

        self.canvas.ui_set_active(True, viewer=self.viewer)

    def pause(self):
        self.canvas.ui_set_active(False)

        # remove the canvas from the image
        p_canvas = self.viewer.get_canvas()
        try:
            p_canvas.delete_object(self.canvas)
        except Exception:
            pass

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None or np.isclose(self._rot_deg, 0.0):
            self.pipeline.send(res_np=data)
            return

        if self.add_alpha:
            minv, maxv = trcalc.get_minmax_dtype(data.dtype)
            data = trcalc.add_alpha(data, alpha=maxv)

        if self.clip:
            res_np = trcalc.rotate_clip(data, self._rot_deg,
                                        logger=self.logger)
        else:
            res_np = trcalc.rotate(data, self._rot_deg, pad=0,
                                   logger=self.logger)

        self.pipeline.send(res_np=res_np)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.rot_deg = d['rot_deg']
        self.clip = d['clip']
        self.add_alpha = d['add_alpha']



"""Unit Tests for some time series helper functions"""

import pytest
import numpy as np
from numpy.testing import assert_allclose

from ginga.plot.data_source import XYDataSource


class TestXYDataSource:

    def setup_class(self):
        self.n = 7
        self.test_pts = [(x, x**2) for x in range(0, self.n)]
        self.test_arr = np.array(self.test_pts, dtype=int)

    def test_initialize_points(self):
        expected = self.test_arr
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=expected)

        actual = dsrc.points
        assert_allclose(actual, expected)

    def test_set_points(self):
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int))

        expected = self.test_arr
        dsrc.set_points(expected)

        actual = dsrc.get_points()
        assert_allclose(actual, expected)

    def test_full_exception(self):
        expected = self.test_arr
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=expected, overwrite=False)

        with pytest.raises(ValueError):
            dsrc.add((8, 8**2))

    def test_is_full(self):
        expected = self.test_arr
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=expected, overwrite=False)

        assert dsrc.is_fullp()

    def test_overwrite(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        to_add = [(8, 8**2), (9, 9**2)]
        for pt in to_add:
            dsrc.add(pt)

        expected = np.array(points[len(to_add):] + to_add, dtype=int)
        actual = dsrc.get_points()
        assert_allclose(actual, expected)

    def test_add_points(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        to_add = [(8, 8**2), (9, 9**2)]
        dsrc.add_points(to_add)

        expected = np.array(points[len(to_add):] + to_add, dtype=int)
        actual = dsrc.get_points()
        assert_allclose(actual, expected)

    def test_peek(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        expected = points[-1]
        actual = dsrc.peek()
        assert_allclose(actual, expected)

    def test_peek_rear(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        expected = points[0]
        actual = dsrc.peek_rear()
        assert_allclose(actual, expected)

    def test_pop(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        expected = points[-1]
        actual = dsrc.pop()
        assert_allclose(actual, expected)

        assert len(dsrc.get_points()) == self.n - 1

    def test_pop_rear(self):
        points = self.test_pts
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=points, overwrite=True)

        expected = points[0]
        actual = dsrc.pop_rear()
        assert_allclose(actual, expected)

        assert len(dsrc.get_points()) == self.n - 1

    def test_points_property(self):
        expected = self.test_arr
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            points=expected, overwrite=True)

        actual = dsrc.points
        assert_allclose(actual, expected)

        assert len(dsrc) == self.n

    def test_limits(self):
        n = 100
        y_arr = np.random.randint(0, high=1000, size=n)
        x_arr = np.arange(0, n)
        points = np.array((x_arr, y_arr)).T
        dsrc = XYDataSource(np.zeros((n, 2), dtype=int),
                            points=points, overwrite=False)

        expected = np.array([(x_arr.min(), y_arr.min()),
                             (x_arr.max(), y_arr.max())])
        actual = dsrc.get_limits()
        assert_allclose(actual, expected)

    def test_empty_exception(self):
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            none_for_empty=False)

        with pytest.raises(ValueError):
            dsrc.peek()

        with pytest.raises(ValueError):
            dsrc.peek_rear()

        with pytest.raises(ValueError):
            dsrc.pop()

        with pytest.raises(ValueError):
            dsrc.pop_rear()

    def test_none_for_empty(self):
        dsrc = XYDataSource(np.zeros((self.n, 2), dtype=int),
                            none_for_empty=True)
        expected = None
        actual = dsrc.peek()
        assert actual == expected

        actual = dsrc.peek_rear()
        assert actual == expected

        actual = dsrc.pop()
        assert actual == expected

        actual = dsrc.pop_rear()
        assert actual == expected



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
``AutoLoad`` is a simple plugin to monitor a folder for new files and
automatically load them into a channel when they appear.

**Plugin Type: Local**

``AutoLoad`` is a local plugin, which means it is associated with a channel.
An instance can be opened for each channel.

.. note:: You need to install the Python "watchdog" package to use this
          plugin.

**Usage**

* To set up a folder to be monitored, type a path of a folder (directory)
  in the "Watched folder" field and press ENTER or click "Set".
* If you need to distinguish between files that will be added to this folder,
  you may type a Python regular expression in the "Regex" box and click "Set".
  Only files with names matching the pattern will be considered.  Note that
  the regex is for the filename only; not any part of the folder path.
* If you ever want to pause the auto loading, you can check the box marked
  "Pause"; this will stop any auto loading.  Note that if you subsequently
  uncheck the box, files that arrived in the intervening period will not be
  loaded.

.. note:: Monitoring folders that reside on network drives may or may not
          work.

**User Configuration**

"""
import os.path
import re

from ginga import GingaPlugin
from ginga.gw import Widgets

# pip install watchdog
have_watchdog = False
try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler
    have_watchdog = True
except ImportError:
    pass

__all__ = ['AutoLoad']


class AutoLoad(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super().__init__(fv, fitsimage)

        # get AutoLoad preferences
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_AutoLoad')
        self.settings.add_defaults(watch_folder=None, filename_regex=None,
                                   start_paused=False)
        self.settings.load(onError='silent')

        self.observer = None
        self._watch = None
        self.data_dir = self.settings.get('watch_folder', None)
        self.regex = self.settings.get('filename_regex', None)
        self.regex_c = None
        self.pause_flag = self.settings.get('start_paused', False)
        self.gui_up = False

    def build_gui(self, container):
        if not have_watchdog:
            raise ImportError("Install 'watchdog' to use this plugin")

        container.set_border_width(2)
        top = Widgets.VBox()
        top.set_border_width(4)

        # Possible future options:
        # - file rename
        # - add metadata
        # - run a script or a custom function
        captions = (("Watched folder:", 'label', 'folder', 'entryset'),
                    ("Match regex:", 'label', 'regex', 'entryset'),
                    ("Pause", 'checkbox'),
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        b.folder.set_tooltip("Folder to monitor for new files")
        if self.data_dir is not None:
            b.folder.set_text(self.data_dir)
        b.folder.add_callback('activated', self.set_folder_cb)
        b.regex.set_tooltip("Regular expression to match file name")
        if self.regex is not None:
            b.folder.set_text(self.regex)
        b.regex.add_callback('activated', self.set_regex_cb)
        b.pause.set_tooltip("Pause auto loading")
        b.pause.set_state(self.pause_flag)
        b.pause.add_callback('activated', self.set_pause_cb)

        top.add_widget(w, stretch=0)
        # for customization by subclasses
        self.w.vbox = top

        container.add_widget(top, stretch=0)

        container.add_widget(Widgets.Label(''), stretch=1)

        btns = Widgets.HBox()
        btns.set_spacing(3)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)
        container.add_widget(btns, stretch=0)

        self.gui_up = True

    def set_folder_cb(self, w):
        data_dir = w.get_text().strip()
        self.set_folder(data_dir)

    def set_folder(self, data_dir):
        if self.data_dir is not None:
            # remove old watch
            if self._watch is not None:
                self.observer.unschedule(self._watch)

        if data_dir is None or len(data_dir) == 0:
            self.data_dir = None
            return
        # add new watch folder
        event_handler = FileSystemEventHandler()
        event_handler.on_closed = self._file_detected_cb
        self._watch = self.observer.schedule(event_handler, data_dir,
                                             recursive=False)
        self.data_dir = data_dir

    def set_regex_cb(self, w):
        regex = w.get_text().strip()
        self.set_regex(regex)

    def set_regex(self, regex):
        if regex is None or len(regex) == 0:
            self.regex_c = None
            self.regex = None
            return

        try:
            self.regex_c = re.compile(regex)
            self.regex = regex

        except Exception as e:
            self.regex = None
            msgerr = "Couldn't compile regular expression: {}".format(e)
            self.fv.show_error(msgerr)

    def set_pause_cb(self, w, tf):
        self.pause_flag = tf

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        if self.regex is not None:
            self.set_regex(self.regex)
        self.observer = Observer()

        if self.data_dir is not None:
            self.set_folder(self.data_dir)

        self.observer.start()

    def stop(self):
        self.observer.stop()
        self._watch = None
        self.observer.join()
        self.gui_up = False

    def _file_detected_cb(self, event):
        if self.pause_flag:
            return
        filepath = event.src_path
        self.logger.info(f"new file detected: {filepath}")

        if self.check_file(filepath):
            self.load_file(filepath)

    def check_file(self, filepath):
        """Check path and return True if we should load it.
        Subclass can override this to add different checks.
        """
        if self.regex is not None:
            fname = os.path.basename(filepath)
            match = self.regex_c.match(fname)
            if not match:
                self.logger.info(f"filename {fname} did not match regex")
                return False

        # file did not fail any checks we set for it
        return True

    def load_file(self, filepath):
        """Subclass can override to change behavior."""
        self.fv.nongui_do(self.fv.load_file, filepath)

    def redo(self):
        pass

    def __str__(self):
        return 'autoload'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_AutoLoad', package='ginga')



"""
Tests of Ginga canvas objects.

"""
import pytest
import logging

import numpy as np

from ginga.pilw.ImageViewPil import CanvasView
from ginga.canvas.CanvasObject import get_canvas_types
from ginga.AstroImage import AstroImage


class TestCanvas(object):

    def setup_class(self):
        self.logger = logging.getLogger("TestCanvas")
        self.viewer = CanvasView(logger=self.logger)
        self.dc = get_canvas_types()
        self.canvas = self.dc.DrawingCanvas()
        # NOTE: canvas needs to be added to a viewer otherwise lots
        # of operations do not work!
        # add our canvas to the viewer's default canvas
        self.viewer.get_canvas().add(self.canvas)

    def test_add_to_canvas(self):
        """Test adding an object to a canvas."""
        o = self.dc.Line(10, 10, -200, -300)
        self.canvas.add(o)
        assert o in self.canvas

    def test_delete_from_canvas(self):
        """Test deleting an object from a canvas."""
        o = self.dc.Line(10, 10, -200, -300)
        self.canvas.add(o)
        assert o in self.canvas
        self.canvas.delete_object(o)
        assert o not in self.canvas

    def test_delete_from_empty_canvas(self):
        """Test deleting an object that does not exist on a canvas."""
        o = self.dc.Line(10, 10, -200, -300)
        with pytest.raises(ValueError) as e:
            self.canvas.delete_object(o)

    def test_clear_canvas(self):
        """Test clearing a canvas."""
        o = self.dc.Line(10, 10, -200, -300)
        self.canvas.add(o)
        o2 = self.dc.Line(-10, -10, 200, 300)
        self.canvas.add(o2)
        self.canvas.delete_all_objects()
        assert o not in self.canvas
        assert o2 not in self.canvas

    def test_copy_simple(self):
        """Test copying a simple object on a canvas."""
        r = self.dc.Rectangle(10, 20, 110, 210)
        self.canvas.add(r)
        r2 = r.copy()
        assert isinstance(r2, self.dc.Rectangle)
        assert r2 is not r

    def test_copy_compound(self):
        """Test copying a compound object on a canvas."""
        a = self.dc.Annulus(100, 200, 10, width=5, atype='circle')
        self.canvas.add(a)
        a2 = a.copy()
        assert isinstance(a2, self.dc.Annulus)
        assert a2 is not a
        # check that children are copied as well, and not shared
        for o1, o2 in zip(a.get_objects(), a2.get_objects()):
            assert o1 is not o2

    def test_move_delta_pt(self):
        """Test moving an object by a delta."""
        o = self.dc.Box(50, 60, 100, 200)
        self.canvas.add(o)
        assert tuple(o.get_center_pt()) == (50, 60)
        o.move_delta_pt((-10, 15))
        assert tuple(o.get_center_pt()) == (40, 75)

    def _setup_color_test(self, obj):
        img = AstroImage(data_np=np.zeros((100, 100), dtype=int))
        self.viewer.set_image(img)
        self.canvas.add(obj)
        mask = img.get_shape_mask(obj)
        data = img.get_data()
        data[mask] = 1
        return np.sum(data)

    def test_color_object_circle(self):
        """Test coloring a circle."""
        res = self._setup_color_test(self.dc.Circle(49.5, 49.5, 12.0))
        assert res == 448

    def test_color_object_box(self):
        """Test coloring a box."""
        res = self._setup_color_test(self.dc.Box(49.5, 49.5, 12.0, 14.0))
        assert res == 672

    def test_color_object_ellipse(self):
        """Test coloring an ellipse."""
        res = self._setup_color_test(self.dc.Ellipse(49.5, 49.5, 12.0, 30.0))
        assert res == 1136

    def test_color_object_line(self):
        """Test coloring a line."""
        res = self._setup_color_test(self.dc.Line(10.0, 10.0, 88.0, 63.0))
        assert res == 196

    def _setup_cutout_test(self, obj):
        img = AstroImage(data_np=np.zeros((100, 100), dtype=int))
        self.viewer.set_image(img)
        self.canvas.add(obj)
        data = img.cutout_shape(obj)
        return data

    def test_cutout_object_circle(self):
        """Test cutting out a circle."""
        data = self._setup_cutout_test(self.dc.Circle(49.5, 49.5, 12.0))
        assert data.shape == (25, 25)
        assert np.ma.count_masked(data) == 177

    def test_cutout_object_box(self):
        """Test cutting out a box."""
        data = self._setup_cutout_test(self.dc.Box(49.5, 49.5, 12.0, 14.0))
        assert data.shape == (29, 25)
        assert np.ma.count_masked(data) == 53

    def test_cutout_object_ellipse(self):
        """Test cutting out an ellipse."""
        data = self._setup_cutout_test(self.dc.Ellipse(49.5, 49.5, 12.0, 30.0))
        assert data.shape == (61, 25)
        assert np.ma.count_masked(data) == 389



#
# syncops.py -- Specialty synchronization classes
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import threading


class BusyError(Exception):
    pass


class Shelf:
    def __init__(self, lock=None):
        if lock is None:
            lock = threading.RLock()
        self._cond = threading.Condition(lock=lock)
        self._is_blocked = False

    def is_blocked(self):
        with self._cond:
            return self._is_blocked

    def block(self):
        with self._cond:
            self._is_blocked = True
        return self

    def unblock(self):
        with self._cond:
            self._is_blocked = False

    def get_stocker(self):
        return Stocker(self)


class Stocker:
    def __init__(self, shelf):
        self.shelf = shelf

    def __enter__(self):
        self.shelf.block()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shelf.unblock()
        return False



#
# Task.py -- Basic command pattern and thread pool implementation.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import time
import threading

import queue as Queue

# NOTE: See http://bugs.python.org/issue7946
# we cannot effectively use threading for loading files/network/etc.
# without setting the switchinterval down on python 3 due to the new
# GIL implementation
_swival = 0.000001
sys.setswitchinterval(_swival)

from . import Callback  # noqa


class TaskError(Exception):
    """Exception generated for task errors"""
    pass


class TaskTimeout(TaskError):
    """Exception generated when timing out waiting on a task"""
    pass


class UserTaskException(Exception):
    pass


# ------------ BASIC TASKS ------------

class Task(Callback.Callbacks):
    """This class implements a basic Task (command) abstraction.  The
    methods define the interface for starting, cancelling, waiting on a
    task, etc.
    """

    def __init__(self):
        """
        The constructor sets bare essentials for a Task object.  See the
        initialize() and start() methods.
        """
        self.ev_done = threading.Event()
        self.tag = None
        self.logger = None
        self.threadPool = None
        # Lock for task state critical sections
        self.tlock = threading.RLock()
        # Parent task can set this (or add to it) explicitly to determine
        # which values will be copied when it calls initialize() on a child
        # task.
        self.shares = ['logger', 'threadPool', 'shares']

        super(Task, self).__init__()

        self.enable_callback('resolved')

    def initialize(self, taskParent, override=None):
        """This method initializes a task for (re)use.  taskParent is the
        object instance of the parent task, or a 'task environment' (something
        that runs tasks).
        If subclass overrides this method, it should call the superclass
        method at some point.

        - Copy shared data from taskParent, overriding items from _override_
          if they are present there ('contagion' of task values).
        - Generate a unique tag, to be used with the Gen2 Monitor.
        - Clear done event, initialize times and result.
        """
        # For now, punt if we have no apparent parent
        if taskParent and hasattr(taskParent, 'shares'):
            # Copy some variables from our parent task, unless they are being
            # overridden explicitly.  Using this general "contagion" mechanism,
            # a task can cause it's children to have values available to them
            # without passing them explicitly.
            for var in taskParent.shares:
                if override and var in override:
                    self.__dict__[var] = override[var]
                else:
                    self.__dict__[var] = taskParent.__dict__[var]

        else:
            #raise TaskError("Cannot initialize task without a taskParent!")
            pass

        # Generate our own unique tag.  'tagger' should have been transmitted
        # from the parent task
        if not self.tag:
            try:
                self.tag = str(taskParent) + '.' + self.tagger.get_tag(self)
            except Exception:
                # Failed--fall back to internal tagger
                self.tag = get_tag(taskParent)

        # Some per-task specific initialization
        self.ev_done.clear()
        self.starttime = time.time()
        self.endtime = 0
        self.totaltime = 0
        self.result = None

        return self.tag

    def start(self):
        """This method starts a task executing and returns immediately.
        Subclass should override this method, if it has an asynchronous
        way to start the task and return immediately.
        """
        if self.threadPool is not None:
            self.threadPool.addTask(self)

            # Lets other threads have a chance to run
            time.sleep(0)
        else:
            raise TaskError("start(): nothing to start for task %s" % self)

    def init_and_start(self, taskParent, override={}):
        """Convenience method to initialize and start a task.
        """
        tag = self.initialize(taskParent, override=override)
        self.start()

        return tag

    def check_state(self):
        """Abstract method that should check for pause, cancellation, or
        any other sort of preemption event.
        """
        pass

    def extend_shares(self, varlist):
        shares = set(self.shares)
        for var in varlist:
            if hasattr(self, var):
                shares.add(var)

        self.shares = shares

    def stop(self):
        """This method cancels an executing task (if possible).
        Subclass should override this method.
        Return True if task could be cancelled, False if not?
        """
        raise TaskError("Task %s: subclass should override stop() method!" % (
                        self))

    def pause(self):
        """This method pauses an executing task (if possible).
        Subclass should override this method.
        Return True if task could be paused, False if not?
        """
        raise TaskError("Task %s: subclass should override pause() method!" % (
                        self))

    def resume(self):
        """This method resumes an executing task (if possible).
        Subclass should override this method, should not call super.resume().
        Return True if task could be resumed, False if not?
        """
        raise TaskError("Task %s: subclass should override resume() method!" % (
            self))

    def wait(self, timeout=None):
        """This method waits for an executing task to finish.
        Subclass can override this method if necessary.
        """
        self.ev_done.wait(timeout=timeout)

        if not self.ev_done.is_set():
            raise TaskTimeout("Task %s timed out." % self)

        # --> self.result is set
        # If it is an exception, then raise it in this waiter
        if isinstance(self.result, Exception):
            raise self.result

        # Release waiters and perform callbacks
        # done() has already been called, because of self.ev_done check
        # "asynchronous" tasks should could call done() here
        #self.done(self.result)

        return self.result

    def step(self):
        """If a task has a way of stepping through an operation.  It can
        implement this method.  Subclass should not call super.step().
        """
        raise TaskError("Task %s: subclass should override step() method!" %
                        self)

    def execute(self):
        """This method does the work of a task (if executed by the
        thread pool) and returns when it is finished.  *** Subclass should
        override this method! ***  It should take no arguments, and can
        return anything.
        """
        raise TaskError("Task %s: subclass should override execute() method!" %
                        self)

    def done(self, result, noraise=False):
        """This method is called when a task has finished executing.
        Subclass can override this method if desired, but should call
        superclass method at the end.
        """
        # [??] Should this be in a critical section?

        # Has done() already been called on this task?
        if self.ev_done.is_set():
            # ??
            if isinstance(self.result, Exception) and (not noraise):
                raise self.result
            return self.result

        # calculate running time and other finalization
        self.endtime = time.time()
        try:
            self.totaltime = self.endtime - self.starttime
        except AttributeError:
            # task was not initialized properly
            self.totaltime = 0.0
        self.result = result

        # Release thread waiters
        self.ev_done.set()

        # Perform callbacks for event-style waiters
        self.make_callback('resolved', self.result)

        # If the result is an exception, then our final act is to raise
        # it in the caller, unless the caller explicitly supressed that
        if isinstance(result, Exception) and (not noraise):
            raise result

        return result

    def get_tag(self):
        """This is only valid AFTER initialize() has been called on the task.
        """
        return self.tag

    def __str__(self):
        """Returns a string representation of a task (e.g. for debugging).
        Subclass can override this method if desired.
        """
        return str(self.tag)

    def __lt__(self, other):
        return False

    def getExecutionTime(self):
        return self.totaltime

    def runTask(self, task, timeout=None):
        """Run a child task to completion.  Returns the result of
        the child task.
        """
        # Initialize the task.
        task.initialize(self)

        # Start the task.
        task.start()

        # Lets other threads run
        time.sleep(0)

        # Wait for it to finish.
        res = task.wait(timeout=timeout)

        # Now we're done
        return res

    def run(self, task, timeout=None):
        """Run a child task to completion.  Returns the result of
        the child task.  Simply calls runTask().
        """
        return self.runTask(task, timeout=timeout)


# For testing...

class printTask(Task):
    """Simple task that prints msg."""
    def __init__(self, msg):
        self.msg = msg
        super(printTask, self).__init__()

    def execute(self):
        print(self.msg)


class sleepTask(Task):
    """Simple task that sleeps for delay seconds."""
    def __init__(self, delay):
        self.delay = delay
        super(sleepTask, self).__init__()

    def execute(self):
        self.ev_done.wait(timeout=self.delay)


class FuncTask(Task):
    """Simple task that calls func and returns func's return value."""
    def __init__(self, func, args, kwdargs, logger=None):
        self.func = func
        self.args = args
        self.kwdargs = kwdargs
        self.logger = logger
        super(FuncTask, self).__init__()

    def execute(self):
        if self.logger:
            # Cap logging size around 500 characters
            s_args = str(self.args)
            if len(s_args) > 500:
                s_args = s_args[:500]
            s_kwdargs = str(self.kwdargs)
            if len(s_kwdargs) > 500:
                s_kwdargs = s_kwdargs[:500]
            self.logger.debug("Running %s(%s, %s)" % (
                self.func.__name__, s_args, s_kwdargs))
            s_args = None
            s_kwdargs = None

        try:
            res = self.func(*self.args, **self.kwdargs)
            self.done(res)

            if self.logger:
                self.logger.debug("Function returned %s" % (
                    str(res)))

        except Exception as e:
            if self.logger:
                self.logger.error("Task '%s' terminated with exception: %s" %
                                  (str(self), str(e)), exc_info=True)
            self.done(e)


class FuncTask2(FuncTask):
    """Simple task that calls func and returns func's return value.
    This version lets you specify the positional and keyword arguments
    more naturally 'in place' in the constructor.
    """
    def __init__(self, func, *args, **kwdargs):
        super(FuncTask2, self).__init__(func, args, kwdargs)

    def set_logger(self, logger):
        self.logger = logger


def make_tasker(func):
    """make_tasker takes a callable (function, method, etc.) and returns
    a new factory function for generating tasks.  Each factory function is
    designed to consume its arguments and return a task that, when executed,
    will call the function upon the arguments.

    TODO: deprecate this and just use FuncTask, which is easier to
    understand--must change a number of programs first.
    """
    def anonFunc(*args, **kwdargs):
        class anonTask(Task):
            def execute(self):
                self.logger.debug("Executing fn %s" % func)
                try:
                    val = func(*args, **kwdargs)

                    self.logger.debug("Done executing fn %s" % func)
                    return val

                except Exception as e:
                    # Log error message and re-raise exception.
                    self.logger.error("fn %s raised exception: %s" % (
                        func, str(e)))
                    raise e

        return anonTask()
    return anonFunc


# ------------ COMPOUND TASKS ------------

class SequentialTaskset(Task):
    """Compound task that runs a series of tasks sequentially.
    """

    def __init__(self, taskseq):

        super(SequentialTaskset, self).__init__()

        self.tasklist = list(taskseq)

    def initialize(self, taskParent, **kwdargs):
        self.index = 0

        super(SequentialTaskset, self).initialize(taskParent, **kwdargs)

    def step(self):
        """Run the next child task and wait for completion (no timeout)."""
        if self.index >= len(self.tasklist):
            raise TaskError("step(): sequential compound task %s finished" % self)

        self.check_state()

        # Select next task from the set and advance the index
        self.task = self.tasklist[self.index]
        self.index += 1

        return self.runTask(self.task)

    def execute(self):
        """Run all child tasks, in order, waiting for completion of each.
        Return the result of the final child task's execution.
        """
        while self.index < len(self.tasklist):
            res = self.step()
            self.logger.debug('SeqSet task %i has completed with result %s' %
                              (self.index, res))

        # Returns result of last task to quit
        return res

    def stop(self):
        """Interrupt/cancel execution, but will allow current child task
        to complete."""
        #self.ev_intr.set()

        try:
            self.task.stop()

        except TaskError as e:
            self.logger.error("Error cancelling child task: %s" % (str(e)))

    def addTask(self, task):
        """Append a task to the task sequence.  If the SequentialTaskset has
        already completed execution, this will do nothing unless it is
        restarted (initialize(), start()).
        """
        self.tasklist.append(task)


class ConcurrentAndTaskset(Task):
    """Compound task that runs a set of tasks concurrently, and does not
    return until they all terminate.
    """

    def __init__(self, taskseq):

        super(ConcurrentAndTaskset, self).__init__()

        self.taskseq = taskseq
        # tuning value for polling inefficiency
        self.idletime = 0.001

        # internal mutex
        self._lock_c = threading.RLock()

    def execute(self):
        """Run all child tasks concurrently in separate threads.
        Return last result after all child tasks have completed execution.
        """

        with self._lock_c:
            self.count = 0
            self.numtasks = 0
            self.taskset = []
            self.results = {}
            self.totaltime = time.time()
            # Start all tasks
            for task in self.taskseq:
                self.taskset.append(task)
                self.numtasks += 1
                task.init_and_start(self)

        num_tasks = self.getNumTasks()
        # Wait on each task to clean up results
        while num_tasks > 0:

            self.check_state()

            for i in range(num_tasks):
                try:
                    try:
                        task = self.getTask(i)
                    except IndexError:
                        # A task got deleted from the set.  Jump back out
                        # to outer loop and repoll the number of tasks
                        break

                    #self.logger.debug("waiting on %s" % task)
                    res = task.wait(timeout=self.idletime)

                    #self.logger.debug("finished: %s" % task)
                    self.child_done(res, task)

                except TaskTimeout:
                    continue

                except Exception as e:
                    #self.logger.warning("Subtask propagated exception: %s" % str(e))
                    self.child_done(e, task)
                    continue

            # wait a bit and try again
            #self.ev_quit.wait(self.idletime)

            # re-get number of tasks, in case some were added or deleted
            num_tasks = self.getNumTasks()

        # Scan results for errors (exceptions) and raise the first one we find
        for key in self.results.keys():
            value = self.results[key]
            if isinstance(value, Exception):
                (count, task) = key
                self.logger.error("Child task %s terminated with exception: %s" % (
                    task.tag, str(value)))
                raise value

        # Return value of last child to complete
        return value

    def child_done(self, result, task):
        with self._lock_c:
            self.count += 1
            self.logger.debug('Concurrent task %d/%d has completed' % (
                self.count, self.numtasks))
            self.taskset.remove(task)
            self.totaltime += task.getExecutionTime()
            self.results[(self.count, task)] = result

    def stop(self):
        """Call stop() on all child tasks, and ignore TaskError exceptions.
        Behavior depends on what the child tasks' stop() method does."""
        with self._lock_c:
            for task in self.taskset:
                try:
                    task.stop()

                except TaskError as e:
                    # Task does not have a way to stop it.
                    # TODO: notify who?
                    pass

        # stop ourself
        #self.ev_intr.set()

    def addTask(self, task):
        """Add a task to the task set.
        """
        # Try to start task first.  If it fails then we don't need to
        # undo adding it to taskset
        task.initialize(self)
        task.start()

        with self._lock_c:
            self.numtasks += 1
            self.taskset.append(task)

    def getTask(self, i):
        with self._lock_c:
            return self.taskset[i]

    def getNumTasks(self):
        """Get the set of active tasks.
        """
        with self._lock_c:
            return len(self.taskset)


class QueueTaskset(Task):
    """Compound task that runs a set of tasks that it reads from a queue
    concurrently.  If _waitflag_ is True, then it will run each task to
    completion before starting the next task.
    """

    def __init__(self, queue, waitflag=True, timeout=0.1, ev_quit=None):

        super(QueueTaskset, self).__init__()

        self.queue = queue
        self.waitflag = waitflag
        self.lock = threading.RLock()
        self.timeout = timeout
        self.task = None
        self.ev_cancel = threading.Event()
        self.ev_pause = threading.Event()
        if ev_quit is None:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit

    def flush(self):
        # Flush queue of pending tasks
        self.logger.debug("Flushing queue.")
        while True:
            try:
                self.queue.get(block=False)
            except Queue.Empty:
                break

    def stop(self):
        self.flush()
        #self.ev_intr.set()

        try:
            if self.task:
                self.task.stop()

        except TaskError as e:
            #self.logger.error("Error cancelling child task: %s" % (str(e)))
            pass

        # put termination sentinel
        self.queue.put(None)

    def stop_child(self):
        self.flush()

        try:
            if self.task:
                self.task.stop()

        except TaskError as e:
            #self.logger.error("Error cancelling child task: %s" % (str(e)))
            pass

    def execute(self):
        self.count = 0
        self.totaltime = 0
        self.logger.debug("Queue Taskset starting")
        while not self.ev_quit.is_set():
            try:
                self.check_state()

                task = self.queue.get(block=True, timeout=self.timeout)
                if task is None:
                    # termination sentinel
                    break

                self.task = task

                task.add_callback('resolved', self.child_done)

                with self.lock:
                    self.count += 1

                self.ev_cancel.clear()
                try:
                    task.initialize(self)

                    self.logger.debug("Starting task '%s'" % str(task))
                    task.start()

                    if self.waitflag:
                        res = task.wait()

                        self.logger.debug("Task %s terminated with result %s" % (
                                          (str(task), str(res))))
                except Exception as e:
                    self.logger.error("Task '%s' terminated with exception: %s" %
                                      (str(task), str(e)), exc_info=True)

                    # If task raised exception then it didn't call done,
                    task.done(e, noraise=True)

            except Queue.Empty:
                # No task available.  Continue trying to get one.
                continue

        # TODO: should we wait for self.count > 0?
        self.logger.debug("Queue Taskset terminating")

        return self.result

    def child_done(self, task, result):
        with self.lock:
            self.count -= 1
            self.totaltime += task.getExecutionTime()
            self.result = result

    def cancel(self):
        self.flush()

        super(QueueTaskset, self).cancel()

    def addTask(self, task):
        self.queue.put(task)


# ------------ PRIORITY QUEUES ------------

class PriorityQueue(Queue.PriorityQueue):
    pass


# ------------ WORKER THREADS ------------

class WorkerThread(object):
    """Container for a thread in which to call the execute() method of a task.
    A WorkerThread object waits on the task queue, executes a task when it
    appears, and repeats.  A call to start() is necessary to start servicing
    the queue, and a call to stop() will terminate the service.
    """

    def __init__(self, queue, logger=None, ev_quit=None,
                 timeout=0.2, tpool=None):

        self.queue = queue
        self.logger = logger
        self.timeout = timeout
        # global termination flag
        if ev_quit is None:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit
        # local termination flag
        self.my_quit = threading.Event()
        self.tpool = tpool
        self.lock = threading.RLock()
        self.thread = None
        self.status = 'stopped'
        self.time_idle = None
        self.tid = None

    def setstatus(self, status):
        """Sets our status field so that others can inquire what we are doing.
        Set of status:
          starting, idle
        """
        with self.lock:
            self.status = status
            if status == 'idle':
                self.time_idle = time.time()
            else:
                self.time_idle = None

    def getstatus(self):
        """Returns our status--a string describing what we are doing.
        """
        with self.lock:
            return (self.status, self.time_idle)

    def execute(self, task):
        """Execute a task.
        """

        taskid = str(task)
        res = None
        try:
            # Try to run the task.  If we catch an exception, then
            # it becomes the result.
            self.setstatus('executing %s' % taskid)

            self.logger.debug("now executing task '%s'" % taskid)
            try:
                res = task.execute()

            except UserTaskException as e:
                res = e

            except Exception as e:
                self.logger.error("Task '%s' raised exception: %s" %
                                  (str(task), str(e)), exc_info=True)
                res = e

        finally:
            self.logger.debug("done executing task '%s'" % str(task))

            self.setstatus('cleaning %s' % taskid)
            # Wake up waiters on other threads
            task.done(res, noraise=True)

            self.setstatus('idle')

    # Basic task execution loop.  Dequeue a task and run it, then look
    # for another one
    def taskloop(self, ev_start):
        self.tid = threading.get_ident()
        self.setstatus('starting')
        self.logger.debug(f"thread {self.tid} starting worker thread loop.")

        # If we were handed a thread pool upon startup, then register
        # ourselves with it.
        if self.tpool is not None:
            self.tpool.register_up(self)

        try:
            self.setstatus('idle')
            if ev_start is not None:
                ev_start.set()
            while not self.ev_quit.is_set() and not self.my_quit.is_set():
                try:
                    # Wait on our queue for a task; will timeout in
                    # self.timeout secs
                    (priority, task) = self.queue.get(block=True,
                                                      timeout=self.timeout)
                    if task is None:
                        # termination sentinel
                        self.queue.put((priority, task))
                        break

                    self.execute(task)

                except Queue.Empty as e:
                    # Reach here when we time out waiting for a task
                    if self.tpool is not None and self.time_idle is not None:
                        idle_sec = time.time() - self.time_idle
                        if (self.tpool.idle_limit_sec is not None and
                            idle_sec > self.tpool.idle_limit_sec):
                            self.tpool.offer_to_quit(self)

        finally:
            if self.tpool is not None:
                self.tpool.register_dn(self)

            self.setstatus('stopped')

        self.logger.debug(f"thread {self.tid} exiting.")

    def start(self, wait=False):
        if self.thread is not None:
            raise RuntimeError("A worker thread is already running")
        self.my_quit.clear()
        ev_start = None
        if wait:
            ev_start = threading.Event()
        self.thread = threading.Thread(target=self.taskloop, args=[ev_start])
        self.thread.start()

        if wait:
            ev_start.wait()

    def stop(self):
        self.my_quit.set()

    def cleanup(self):
        if self.thread is not None:
            alive = self.thread.is_alive()
            if not alive:
                self.thread.join()
            self.thread = None


# ------------ THREAD POOL ------------

class ThreadPool(object):
    """A simple thread pool for executing tasks asynchronously.

    self.status states:
        down    no threads are ready for service
        up      all threads are ready for service
        start   threads are starting, but not all of them are up yet
        stop    threads are stopping, but not all of them are down yet
    """

    def __init__(self, numthreads=1, logger=None, ev_quit=None,
                 minthreads=None, idle_limit_sec=10.0,
                 workerClass=WorkerThread, analyze_interval=None):

        self.numthreads = max(1, numthreads)
        self.logger = logger
        if ev_quit is None:
            ev_quit = threading.Event()
        self.ev_quit = ev_quit
        self.lock = threading.RLock()
        self.workerClass = workerClass
        if minthreads is None:
            minthreads = numthreads
        self.minthreads = max(0, minthreads)
        self.idle_limit_sec = idle_limit_sec
        self.mon_thread = None
        self._analyze_time = 0.0
        self.analyze_interval = analyze_interval

        self.queue = PriorityQueue()

        # Used to synchronize thread pool startup (see register() method)
        self.regcond = threading.Condition()
        self.mp_cond = threading.Condition()
        self.status = 'down'
        self.waiting = [self.workerClass(self.queue, logger=self.logger,
                                         ev_quit=self.ev_quit, tpool=self)
                        for i in range(self.numthreads)]
        self.running = []
        self.cleanup = []

    def startall(self, wait=False, **kwargs):
        """Start all of the threads in the thread pool.  If _wait_ is True
        then don't return until all threads are up and running.  Any extra
        keyword arguments are passed to the worker thread constructor.
        """
        if self.mon_thread is not None:
            self.logger.error("ignoring duplicate request to start thread pool")
            return

        self.logger.debug("startall called, starting pool attendant thread")
        self.status = 'start'
        self.mon_thread = threading.Thread(target=self.pool_attendant, args=[])
        self.mon_thread.start()

        # if started with wait=True, then expect that threads will register
        # themselves and last one up will set status to "up"
        if wait:
            with self.regcond:
                # Threads are on the way up.  Wait until last one starts.
                while self.status != 'up' and not self.ev_quit.is_set():
                    self.logger.debug("waiting for threads: count=%d" %
                                      len(self.running))
                    self.regcond.wait()
        self.logger.debug("startall done")

    def stopall(self, wait=False):
        """Stop all threads in the worker pool.  If _wait_ is True
        then don't return until all threads are down.
        """
        self.logger.debug("stopall called")
        with self.lock:
            self.status = 'stop'
        # Signal to all threads to terminate.
        self.ev_quit.set()

        if wait:
            with self.regcond:
                # Threads are on the way down.  Wait until last one quits.
                while self.status != 'down':
                    self.logger.debug("waiting for threads: count=%d" %
                                      len(self.running))
                    self.regcond.wait()

        self.mon_thread.join()
        self.mon_thread = None
        self.logger.debug("stopall done")

    def add_threads(self, add_numthreads, minthreads=None):
        with self.regcond:
            self.waiting.extend([self.workerClass(self.queue,
                                                  logger=self.logger,
                                                  ev_quit=self.ev_quit,
                                                  tpool=self)
                                 for i in range(add_numthreads)])
            self.numthreads += add_numthreads
            if minthreads is not None:
                self.minthreads = max(0, minthreads)

    def workerStatus(self):
        with self.regcond:
            return list(map(lambda t: t.getstatus(), self.running))

    def addTask(self, task, priority=0):
        """Add a task to the queue of tasks.

        The task will be executed in a worker thread as soon as one is available.
        Tasks are executed in first-come-first-served order.
        """
        self.queue.put((priority, task))
        with self.mp_cond:
            # wake up pool attendant thread to check on things
            self.mp_cond.notify()

    def pool_attendant(self):
        """Monitor the thread pool as the "pool attendant".

        A thread is started in this method to monitor the thread pool and
        clean up or activate new threads as needed.
        """
        self.logger.debug("starting the thread pool attendant loop...")
        while not self.ev_quit.is_set():
            if self.analyze_interval is not None:
                cur_time = time.time()
                if cur_time - self._analyze_time > self.analyze_interval:
                    self._analyze_time = cur_time
                    self.analyze_threads()

            worker = None
            with self.regcond:
                # join threads that have exited
                while len(self.cleanup) > 0:
                    dead_worker = self.cleanup.pop()
                    dead_worker.cleanup()

                num_running = len(self.running)
                if (num_running < self.minthreads or
                    self.queue.qsize() > 0 and num_running < self.numthreads):
                    assert (len(self.waiting) > 0)
                    worker = self.waiting[0]

            if worker is not None:
                worker.start(wait=True)
            else:
                with self.mp_cond:
                    self.mp_cond.wait(timeout=0.25)

        self.logger.debug("stopping the thread pool attendant loop...")

    def analyze_threads(self):
        self.logger.info("--- analyzing active threads...")
        count = 0
        for thread in threading.enumerate():
            count += 1
            if thread.ident is None:
                # Exclude threads that haven't started yet
                self.logger.info(f"{count:3d}: thread named {thread.name} is initializing...")
                continue

            self.logger.info(f"{count:3d}: thread name: {thread.name}, thread id: {thread.ident}")
            try:
                # Get the top-level stack frame for the thread
                frame = sys._current_frames().get(thread.ident)
                if frame:
                    # Iterate through the stack frames to find the current function
                    while frame:
                        function_name = frame.f_code.co_name
                        self.logger.info(f"currently in function: {function_name}")
                        frame = frame.f_back   # Move to the calling frame
                        if function_name != '<module>':
                            # Stop if we're not in the global scope
                            break
                else:
                    self.logger.warning("could not retrieve stack frame for this thread (might be idle or finished).")
            except Exception as e:
                self.logger.error(f"error inspecting thread {thread.name}: {e}")
        self.logger.info("--- done analyzing")

    def delTask(self, taskid):
        self.logger.error("delTask not yet implemented")

    def purgeTasks(self):
        self.logger.error("purgeTasks not yet implemented")

    def offer_to_quit(self, worker):
        """Called by WorkerThread objects when they have been idle
        for a certain period.
        """
        with self.regcond:
            if len(self.running) <= self.minthreads or self.queue.qsize() > 0:
                return

            worker.stop()

    def register_up(self, worker):
        """Called by WorkerThread objects to register themselves.

        Acquire the condition variable for the WorkerThread objects.
        Increment the running-thread count.  If we are the last thread to
        start, set status to 'up'.  This allows startall() to complete
        if it was called with wait=True.
        """
        with self.regcond:
            self.waiting.remove(worker)
            self.running.append(worker)
            num_running = len(self.running)
            self.logger.debug("register_up: (%d) count is %d" % (
                worker.tid, num_running))
            if num_running == self.minthreads:
                self.status = 'up'
                self.regcond.notify()

    def register_dn(self, worker):
        """Called by WorkerThread objects to de-register themselves.

        Acquire the condition variable for the WorkerThread objects.
        Decrement the running-thread count.  If we are the last thread to
        start, release the ThreadPool thread, which is stuck in start()
        """
        with self.regcond:
            self.running.remove(worker)
            self.waiting.append(worker)
            self.cleanup.append(worker)
            num_running = len(self.running)
            self.logger.debug("register_dn: (%d) count is %d" % (
                worker.tid, num_running))
            with self.mp_cond:
                # wake up pool attendant to clean up thread
                self.mp_cond.notify()
            if num_running == 0:
                self.status = 'down'
                self.regcond.notify()

    # TO BE DEPRECATED
    addThreads = add_threads


# ------------ SUPPORT FUNCTIONS ------------

_lock_seqnum = threading.Lock()
_count_seqnum = 0


def get_tag(taskParent):
    global _count_seqnum
    with _lock_seqnum:
        generic_id = 'task%d' % (_count_seqnum)
        _count_seqnum += 1

    if taskParent:
        tag = str(taskParent) + '.' + generic_id
    else:
        tag = generic_id

    return tag


# END



#
# plot_mode.py -- mode for manipulating plots
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.ImageView import ImageViewBase
from ginga.modes.mode_base import Mode


class PlotMode(Mode):
    """
    Plot Mode enables bindings for manipulating plots.

    Enter the mode by
    -----------------
    Mode is set by the PlotAide

    Exit the mode by
    ----------------
    * Esc (but this is not recommended)

    Default bindings in mode
    ------------------------
    * x : toggle autoaxis X between ON and OFF
    * y : toggle autoaxis Y between ON and OFF
    * v : toggle autoaxis Y between VIS and OFF
    * p : toggle autoaxis X between PAN and OFF
    * scroll : zoom (scale) the plot in X
    * Ctrl + scroll : zoom (scale) the plot in Y

    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        # return (isinstance(viewer, ImageViewBase) and
        #         # verify that we have a plot aide set
        #         getattr(viewer, 'aide', None) is not None)
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        # this get set to the associated PlotAide object
        self.aide = None

        self.actions = dict(
            dmod_plot=['__^', None, None],

            # NOTE: these only here to generate events. The events are
            # handled by the PlotAide object
            kp_plot_autoaxis_x=['plot+x'],
            kp_plot_autoaxis_y=['plot+y'],
            kp_plot_visible_y=['plot+v'],
            kp_plot_position_x=['plot+p'],

            sc_plot_zoom=['plot+scroll', 'plot+ctrl+scroll'],
            sc_plot_zoom_pan=['plot+pan', 'plot+ctrl+pan'])

    def __str__(self):
        return 'plot'

    def set_aide_ref(self, aide):
        self.aide = aide

    def assert_has_aide(self):
        if self.aide is None:
            raise ValueError("This viewer does not have a plot aide registered")

    def start(self):
        #self.viewer.switch_cursor('plot')
        pass

    def stop(self):
        #self.viewer.switch_cursor('pick')
        self.onscreen_message(None)

    def kp_plot_autoaxis_y(self, *args):
        """Callback invoked when the user presses 'y' in the viewer window.
        """
        self.assert_has_aide()
        autoaxis_y = self.aide.settings['autoaxis_y'] == 'on'
        self.aide.settings['autoaxis_y'] = 'off' if autoaxis_y else 'on'

    def kp_plot_visible_y(self, *args):
        """Callback invoked when the user presses 'v' in the viewer window.
        """
        self.assert_has_aide()
        autoaxis_y = self.aide.settings['autoaxis_y'] == 'vis'
        self.aide.settings['autoaxis_y'] = 'off' if autoaxis_y else 'vis'

    def kp_plot_autoaxis_x(self, *args):
        """Callback invoked when the user presses 'x' in the viewer window.
        """
        self.assert_has_aide()
        autoaxis_x = self.aide.settings['autoaxis_x'] == 'on'
        self.aide.settings['autoaxis_x'] = 'off' if autoaxis_x else 'on'

    def kp_plot_position_x(self, *args):
        """Callback invoked when the user presses 'p' in the viewer window.
        """
        self.assert_has_aide()
        autopan_x = self.aide.settings['autoaxis_x'] == 'pan'
        self.aide.settings['autoaxis_x'] = 'off' if autopan_x else 'pan'

    def sc_plot_zoom(self, viewer, event):
        """Callback called when the user scrolls in the viewer window.
        From this we generate a 'plot-zoom-x' or a 'plot-zoom-y' event.
        """
        self.assert_has_aide()
        bd = viewer.get_bindings()
        direction = bd.get_direction(event.direction)
        zoom_direction = 'in' if direction == 'up' else 'out'
        # default is to zoom the X axis unless CTRL is held down
        zoom_axis = 'y' if 'ctrl' in event.modifiers else 'x'
        event = 'plot-zoom-{}'.format(zoom_axis)
        # turn this into a zoom event for any callbacks registered for it
        self.aide.make_callback(event, zoom_direction)
        return True

    def scroll_plot_zoom_pan(self, viewer, event):
        """Callback called when the user pans in the viewer window
        (i.e. touchpad pan event); we turn this into a zoom event.
        """
        event = self._pa_synth_scroll_event(event)
        if event.state != 'move':
            return False
        self.scroll_plot_zoom(viewer, event)
        return True



#
# CanvasRenderMpl.py -- for rendering into a ImageViewMpl widget
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import matplotlib.patches as patches
import matplotlib.lines as lines
from matplotlib.path import Path as MplPath
#import matplotlib.patheffects as path_effects

from . import MplHelp
from ginga.canvas.mixins import *  # noqa
from ginga.canvas import render
# force registration of all canvas types
import ginga.canvas.types.all  # noqa
from ginga import trcalc


class RenderContext(render.RenderContextBase):

    def __init__(self, renderer, viewer, surface):
        render.RenderContextBase.__init__(self, renderer, viewer)
        self.shape = None

        # TODO: encapsulate this drawable
        self.ctx = MplHelp.MplContext(surface)

        # special scaling for Mpl text drawing to normalize it relative
        # to other backends
        self._font_scale_factor = 1.5

    def text_extents(self, text, font=None):
        if font is None:
            font = self.font
        return self.ctx.text_extents(text, font)

    ##### DRAWING OPERATIONS #####

    def draw_image(self, cvs_img, cpoints, rgb_arr, whence, order='RGBA'):
        return

    def draw_text(self, cx, cy, text, rot_deg=0.0, font=None, fill=None,
                  line=None):
        fontdict = self.ctx.get_fontdict(font, fill)
        self.ctx.push(allow=['alpha', 'color'])
        self.ctx.set(rotation=rot_deg)

        text = self.ctx.axes.text(cx, cy, text, fontdict=fontdict,
                                  **self.ctx.kwdargs)
        # See: https://github.com/matplotlib/matplotlib/issues/7227/
        # text.set_path_effects([path_effects.Stroke(linewidth=self.pen.linewidth,
        #                                            foreground='black'),
        #                        path_effects.Normal()])
        self.ctx.pop()

    def draw_polygon(self, cpoints, line=None, fill=None):
        self.ctx.init(closed=True, transform=None)
        self.ctx.update_patch(line, fill)

        xy = trcalc.strip_z(cpoints)

        p = patches.Polygon(xy, **self.ctx.kwdargs)
        self.ctx.axes.add_patch(p)

    def draw_circle(self, cx, cy, cradius, line=None, fill=None):
        self.ctx.init(radius=cradius, transform=None)
        self.ctx.update_patch(line, fill)

        xy = (cx, cy)

        p = patches.Circle(xy, **self.ctx.kwdargs)
        self.ctx.axes.add_patch(p)

    def draw_bezier_curve(self, verts, line=None):
        self.ctx.init(transform=None)
        self.ctx.update_patch(line, None)

        codes = [MplPath.MOVETO,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 ]
        path = MplPath(verts, codes)

        p = patches.PathPatch(path, **self.ctx.kwdargs)
        self.ctx.axes.add_patch(p)

    def draw_ellipse_bezier(self, verts, line=None, fill=None):
        self.ctx.init(transform=None)
        self.ctx.update_patch(line, fill)

        # draw 4 bezier curves to make the ellipse
        codes = [MplPath.MOVETO,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 MplPath.CURVE4, MplPath.CURVE4, MplPath.CURVE4,
                 ]
        path = MplPath(verts, codes)

        p = patches.PathPatch(path, **self.ctx.kwdargs)
        self.ctx.axes.add_patch(p)

    def draw_line(self, cx1, cy1, cx2, cy2, line=None):
        self.ctx.init(transform=None)
        self.ctx.update_line(line)

        l = lines.Line2D((cx1, cx2), (cy1, cy2), **self.ctx.kwdargs)
        self.ctx.axes.add_line(l)

    def draw_path(self, cpoints, line=None):
        self.ctx.init(closed=False, transform=None)
        self.ctx.update_patch(line, None)

        xy = trcalc.strip_z(cpoints)

        p = patches.Polygon(xy, **self.ctx.kwdargs)
        self.ctx.axes.add_patch(p)


class CanvasRenderer(render.StandardPipelineRenderer):

    def __init__(self, viewer):
        render.StandardPipelineRenderer.__init__(self, viewer)

        self.kind = 'mpl'
        self.rgb_order = viewer.rgb_order
        self.surface = None

    def resize(self, dims):
        """Resize our drawing area to encompass a space defined by the
        given dimensions.
        """
        super(CanvasRenderer, self).resize(dims)

    def render_image(self, data, order, win_coord):
        # for compatibility with the other renderers
        return self.viewer.render_image(data, order, win_coord)

    def get_surface_as_array(self, order=None):
        raise render.RenderError("This renderer can only be used with a matplotlib viewer")

    def setup_cr(self, shape):
        cr = RenderContext(self, self.viewer, self.viewer.ax_util)
        return cr

    def get_dimensions(self, shape):
        cr = self.setup_cr(shape)
        font = cr.get_font_from_shape(shape)
        return cr.text_extents(shape.text, font=font)

    def text_extents(self, text, font):
        cr = RenderContext(self, self.viewer, self.viewer.ax_util)
        return cr.text_extents(text, font=font)


#END



#
# GingaPlugin.py -- Base classes for plugins in Ginga reference viewer
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.misc import Bunch
from ginga.gw import Widgets

__all__ = ['PluginError', 'GlobalPlugin', 'LocalPlugin',
           'ParentPlugin', 'ChildPlugin']


class PluginError(Exception):
    """Plugin related error."""
    pass


class BasePlugin(object):
    """Base class for all plugins."""
    def __init__(self, fv, ident=None):
        super().__init__()
        self.fv = fv
        self.logger = fv.logger
        self.ident = ident

        # Holds GUI widgets
        self.w = Bunch.Bunch()

    def __str__(self):
        if self.ident is not None:
            return self.ident
        return super().__str__()

    # def build_gui(self, container):
    #     """
    #     If a plugin defines this method, it will be called with a
    #     container object in which to build its GUI. It should finish
    #     by packing into this container.  This will be called every
    #     time the plugin is activated.
    #     """
    #     pass

    def start(self):
        """
        This method is called to start the plugin.
        It is called after build_gui().
        """
        pass

    def stop(self):
        """This method is called to stop the plugin."""
        pass

    def _get_docstring(self):
        import inspect

        # Insert section title at the beginning
        plg_name = self.__class__.__name__
        plg_mod = inspect.getmodule(self)
        plg_doc = ('{}\n{}\n'.format(plg_name, '=' * len(plg_name)) +
                   plg_mod.__doc__)
        return plg_name, plg_doc

    def _help_docstring(self):
        plg_name, plg_doc = self._get_docstring()
        self.fv.help_text(plg_name, plg_doc, text_kind='rst', trim_pfx=4)

    def help(self, text_kind='rst'):
        """Display help for the plugin."""
        self.fv.help_plugin(self, text_kind=text_kind)


class GlobalPlugin(BasePlugin):
    """Class to handle a global plugin."""
    def __init__(self, fv, ident=None):
        super().__init__(fv, ident=ident)

    def handleable(self, dataobj):
        """Test whether `dataobj` can be handled by this plugin."""
        # should ideally be overridden by subclass
        return True

    # def redo(self, channel, dataobj):
    #     """If defined, this method is called when a data object is switched to
    #     in a channel."""
    #     pass

    # def blank(self, channel):
    #     """If defined, this method is called when a channel is no longer
    #     displaying any object."""
    #     pass


class LocalPlugin(BasePlugin):
    """Class to handle a local plugin."""
    def __init__(self, fv, fitsimage, ident=None):
        super().__init__(fv, ident=ident)
        self.fitsimage = fitsimage

        # find our channel info
        if self.fitsimage is not None:
            self.chname = self.fv.get_channel_name(self.fitsimage)
            self.channel = self.fv.get_channel(self.chname)
            # TO BE DEPRECATED
            self.chinfo = self.channel

    def handleable(self, dataobj):
        """Test whether `dataobj` can be handled by this plugin."""
        # should ideally be overridden by subclass
        return True

    def modes_off(self):
        """Turn off any mode user may be in."""
        bm = self.fitsimage.get_bindmap()
        bm.reset_mode(self.fitsimage)

    def pause(self):
        """
        This method is called when the plugin is defocused.
        The plugin should disable any user input that it responds to.
        """
        pass

    def resume(self):
        """
        This method is called when the plugin is focused.
        The plugin should enable any user input that it responds to.
        """
        pass

    # def redo(self):
    #     """
    #     If defined, this method is called when a data object is switched
    #     to in the channel associated with the plugin.  It can optionally
    #     redo whatever operation it is doing.
    #     """
    #     pass

    # def blank(self):
    #     """
    #     If defined, this method is called when no object is displayed in the
    #     channel associated with the plugin.  It can optionally clear whatever
    #     operation it is doing.
    #     """
    #     pass


class ParentPlugin(GlobalPlugin):

    def __init__(self, fv):
        # superclass defines some variables for us, like logger
        super().__init__(fv, ident=None)

        self.plugin_dct = dict()
        self.class_childplugin = None

        fv.add_callback('add-channel', self.add_channel)
        fv.add_callback('delete-channel', self.delete_channel)
        fv.add_callback('channel-change', self.focus_cb)
        self.gui_up = False

    def build_gui(self, container):
        vbox = Widgets.VBox()
        vbox.set_border_width(1)
        vbox.set_spacing(1)
        self.w.top_w = vbox

        nb = Widgets.StackWidget()
        vbox.add_widget(nb, stretch=1)
        self.w.nb = nb

        container.add_widget(vbox, stretch=1)
        self.gui_up = True

    def add_channel(self, viewer, channel):
        if not self.gui_up:
            return

        # create child plugin
        plugin = self.class_childplugin(self.fv, channel.fitsimage, self)

        if hasattr(plugin, 'build_gui'):
            # build it's gui
            vbox = Widgets.VBox()
            vbox.set_border_width(1)
            vbox.set_spacing(1)

            plugin.build_gui(vbox)

            self.w.nb.add_widget(vbox, title=channel.name)
        else:
            vbox = None

        self.plugin_dct[channel.name] = Bunch.Bunch(plugin=plugin,
                                                    widget=vbox)

    def delete_channel(self, viewer, channel):
        if not self.gui_up:
            return
        chname = channel.name
        self.logger.debug("deleting channel %s" % (chname))
        bnch = self.plugin_dct[chname]
        del self.plugin_dct[chname]
        widget = bnch.widget
        try:
            bnch.plugin.stop()
        except Exception as e:
            self.logger.error(f"error closing plugin: {e}")
        if widget is not None:
            self.w.nb.remove(widget, delete=True)

    def focus_cb(self, viewer, channel):
        if not self.gui_up:
            return
        bnch = self.plugin_dct[channel.name]
        if bnch.widget is not None:
            index = self.w.nb.index_of(bnch.widget)
            self.w.nb.set_index(index)

    def start(self):
        names = self.fv.get_channel_names()
        for name in names:
            channel = self.fv.get_channel(name)
            self.add_channel(self.fv, channel)

        channel = self.fv.get_channel_info()
        if channel is not None:
            viewer = channel.fitsimage

            image = viewer.get_image()
            if image is not None:
                self.redo(channel, image)

            self.focus_cb(viewer, channel)

    def stop(self):
        names = self.fv.get_channel_names()
        for name in names:
            channel = self.fv.get_channel(name)
            self.delete_channel(self.fv, channel)

        self.gui_up = False
        # dereference gui widgets
        self.w = Bunch.Bunch()
        return True

    def close(self):
        self.fv.stop_global_plugin(str(self))
        return True

    def redo(self, channel, image):
        bnch = self.plugin_dct[channel.name]
        bnch.plugin.redo()


class ChildPlugin(LocalPlugin):

    def __init__(self, fv, image_viewer, parent_plugin):
        super().__init__(fv, image_viewer)

        self.pp = parent_plugin
        self.gui_up = False


# END



import hashlib
import numpy as np
import pytest

zarr = pytest.importorskip('zarr')

from ginga import AstroImage, trcalc
from ginga.misc import log


class TestZarr:
    def setup_class(self):
        self.logger = log.get_logger("TestZarr", null=True)

    def _getdata(self, shape, data_np=None):
        if data_np is None:
            data_np = np.min(np.indices(shape), axis=0)
        data_np = data_np.reshape(shape)
        data_z = zarr.creation.array(data_np, chunks=10)
        return data_z

    def test_zarr_slice_trcalc(self):
        """Test that we can get a subslice of a zarr.
        """
        arr_z = self._getdata((1000, 500))

        x_slice, y_slice = slice(12, 499, 3), slice(10, 951, 11)
        view = (y_slice, x_slice)
        data_np = trcalc.fancy_index(arr_z, view)
        assert isinstance(data_np, np.ndarray)
        assert data_np.shape == (86, 163)
        assert isinstance(data_np[0, 0], np.integer)
        res = '177e1ed261ea24df277511078631ec0f95dfc3e781ac15b2d200f0f0040282ae'
        m = hashlib.sha256()
        m.update(str(data_np.tolist()).encode())
        assert m.hexdigest() == res

    def test_zarr_slice_aimg(self):
        """Test that we can get a subslice of an AstroImage object.
        """
        aimg = AstroImage.AstroImage(logger=self.logger)
        aimg.set_data(self._getdata((700, 800)))

        x_slice, y_slice = slice(12, 800, 8), slice(0, 700, 10)
        view = (y_slice, x_slice)
        data_np = aimg._slice(view)
        assert isinstance(data_np, np.ndarray)
        assert data_np.shape == (70, 99)
        assert isinstance(data_np[0, 0], np.integer)
        res = 'd6f0e61dc54f0c888c8f79d94ead85f8d3c4736efede289ff9946e0091960524'
        m = hashlib.sha256()
        m.update(str(data_np.tolist()).encode())
        assert m.hexdigest() == res

    def test_zarr_aimg_get_data_xy(self):
        """Test that we can get a single value from an AstroImage object.
        """
        aimg = AstroImage.AstroImage(logger=self.logger)
        aimg.set_data(self._getdata((5, 5), data_np=np.arange(0, 25)))

        val = int(aimg.get_data_xy(3, 3))
        assert isinstance(val, int)
        assert val == 18

    def test_zarr_fancy_scale(self):
        """Test that we can get a fancy superslice of a zarr.
        """
        arr_z = self._getdata((5, 5, 5))

        p1 = (0, 0, 0)
        p2 = (5, 5, 5)
        new_dims = (51, 51, 51)

        data_np, scales = trcalc.get_scaled_cutout_wdhtdp(arr_z, p1, p2,
                                                          new_dims,
                                                          logger=self.logger)
        assert isinstance(data_np, np.ndarray)
        assert data_np.shape == new_dims
        assert isinstance(data_np[0, 0, 0], np.integer)
        res = '4d6bb43463f435d76d226c38314fa22a5ba540b7db785b1ccfd2c75d84063fc4'
        m = hashlib.sha256()
        m.update(str(data_np.tolist()).encode())
        assert m.hexdigest() == res



#
# dist.py -- mode for controlling color distribution
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga.ImageView import ImageViewBase
from ginga.modes.mode_base import Mode


class DistMode(Mode):
    """
    Dist Mode enables bindings that can adjust the color distribution
    of an image in a Ginga image viewer.

    These algorithms are similar to "curves" type profiles: "linear",
    "log", "power", "sqrt", "squared", "asinh", "sinh", "histeq"

    Enter the mode by
    -----------------
    * Space, then "d"

    Exit the mode by
    ----------------
    * Esc

    Default bindings in mode
    ------------------------
    * D : reset the color distribution algorithm to "linear"
    * b, up arrow : select the previous distribution algorithm in the list
    * n, down arrow : select the next distribution algorithm in the list
    * scroll wheel : select the color distribution algorithm by scrolling
    * pan gesture : select the color distribution algorithm by swiping
      (hint: finalize selection of algorithm with up/down arrow keys)
    """

    @classmethod
    def is_compatible_viewer(cls, viewer):
        return isinstance(viewer, ImageViewBase)

    def __init__(self, viewer, settings=None):
        super().__init__(viewer, settings=settings)

        self.actions = dict(
            dmod_dist=['__d', None, None],

            kp_dist_reset=['D', 'dist+d', 'dist+D'],
            kp_dist_prev=['dist+up', 'dist+b'],
            kp_dist_next=['dist+down', 'dist+n'],

            sc_dist=['dist+scroll'],

            pa_dist_cycle=['dist+pan'])

    def __str__(self):
        return 'dist'

    @property
    def cancmap(self):
        bd = self.viewer.get_bindings()
        return bd.get_feature_allow('cmap')

    def start(self):
        pass

    def stop(self):
        self.onscreen_message(None)

    def _cycle_dist(self, viewer, msg, direction='down'):
        msg = self.settings.get('msg_dist', msg)
        rgbmap = viewer.get_rgbmap()
        algs = rgbmap.get_hash_algorithms()
        algname = rgbmap.get_hash_algorithm()
        idx = algs.index(algname)
        if direction == 'down':
            idx = (idx + 1) % len(algs)
        else:
            idx = idx - 1
            if idx < 0:
                idx = len(algs) - 1
        algname = algs[idx]
        rgbmap.set_hash_algorithm(algname)
        if msg:
            self.onscreen_message("Color dist: %s" % (algname),
                                  delay=1.0)

    def _reset_dist(self, viewer, msg):
        msg = self.settings.get('msg_dist', msg)
        rgbmap = viewer.get_rgbmap()
        algname = 'linear'
        rgbmap.set_hash_algorithm(algname)
        if msg:
            self.onscreen_message("Color dist: %s" % (algname),
                                  delay=1.0)

    #####  KEYBOARD ACTION CALLBACKS #####

    def kp_dist(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_dist(viewer, msg)

    def kp_dist_reset(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._reset_dist(viewer, msg)

    def kp_dist_prev(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_dist(viewer, msg, direction='up')

    def kp_dist_next(self, viewer, event, data_x, data_y, msg=True):
        if not self.cancmap:
            return False
        event.accept()
        self._cycle_dist(viewer, msg, direction='down')

    #####  SCROLL ACTION CALLBACKS #####

    def sc_dist(self, viewer, event, msg=True):
        """Interactively change the color distribution algorithm
        by scrolling.
        """
        if not self.cancmap:
            return False
        event.accept()
        direction = self.get_direction(event.direction)
        self._cycle_dist(viewer, msg, direction=direction)

    ##### GESTURE ACTION CALLBACKS #####

    def pa_dist_cycle(self, viewer, event, msg=True):
        """Change the color distribution algorithm by a pan gesture.
        (the back end must support gestures)
        """
        if not self.cancmap:
            return False
        event.accept()
        event = self._pa_synth_scroll_event(event)
        if event.state == 'move':
            rev = self.settings.get('zoom_scroll_reverse', False)
            direction = self.get_direction(event.direction, rev=rev)
            self._cycle_dist(viewer, msg, direction=direction)

            return True
        return False



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.canvas.CanvasObject import get_canvas_types
#from ginga import trcalc
from ginga.gw import Widgets

from .base import Stage, StageAction


class Crop(Stage):

    _stagename = 'crop-image'

    def __init__(self):
        super().__init__()

        self.dc = get_canvas_types()
        self.cropcolor = 'yellow'
        self._crop_rect = (0.0, 0.0, 1.0, 1.0)
        self._aspect = None
        self._img_dims = (1, 1)

        canvas = self.dc.DrawingCanvas()
        canvas.enable_edit(True)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.set_draw_mode('edit')
        self.canvas = canvas

    def build_gui(self, container):
        self.viewer = self.pipeline.get('viewer')

        fr = Widgets.Frame("Crop")

        captions = (('Crop %:', 'label', 'crop', 'llabel'),
                    ('Output size:', 'label', 'size', 'llabel'),
                    ('Aspect:', 'label', 'aspect', 'entryset'),
                    )
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        arr = np.asarray(self._crop_rect) * 100.0
        crop = "%6.2f,%6.2f to %6.2f,%6.2f" % tuple(arr)
        b.crop.set_text(crop)

        b.aspect.set_tooltip("Set the aspect ratio (wd/ht)")
        b.aspect.add_callback('activated', self._set_aspect_cb)
        if self._aspect is not None:
            b.aspect.set_text(str(self._aspect))

        fr.set_widget(w)
        container.set_widget(fr)

        self.canvas.set_surface(self.viewer)
        self.canvas.register_for_cursor_drawing(self.viewer)

        wd, ht = 100, 100
        self.crop_obj = self.dc.CompoundObject(
            self.dc.Rectangle(0, 0, wd, ht,
                              color=self.cropcolor),
            self.dc.Text(0, 0, "Crop",
                         color=self.cropcolor))
        self.crop_obj.objects[1].editable = False

        self._gui_update_crop()
        self.w.size.set_text("unknown")

    @property
    def crop_rect(self):
        return self._crop_rect

    @crop_rect.setter
    def crop_rect(self, val):
        self._crop_rect = val
        if self.gui_up:
            self._gui_update_crop()

    @property
    def aspect(self):
        return self._aspect

    @aspect.setter
    def aspect(self, val):
        self._aspect = val
        if self.gui_up:
            asp = self._aspect
            self.w.aspect.set_text('' if asp is None else str(asp))

    def _gui_update_crop(self):
        arr = np.asarray(self._crop_rect) * 100.0
        crop = "%6.2f,%6.2f to %6.2f,%6.2f" % tuple(arr)

        self.w.crop.set_text(crop)

    def resume(self):
        # insert canvas, if not already
        p_canvas = self.viewer.get_canvas()
        if not p_canvas.has_object(self.canvas):
            p_canvas.add(self.canvas)

        if not self.canvas.has_object(self.crop_obj):
            self.canvas.add(self.crop_obj)

        self.canvas.ui_set_active(True, viewer=self.viewer)

    def pause(self):
        self.canvas.ui_set_active(False)

        # remove the canvas from the image
        p_canvas = self.viewer.get_canvas()
        try:
            p_canvas.delete_object(self.canvas)
        except Exception:
            pass

    def edit_cb(self, canvas, obj):
        if obj.kind != 'rectangle':
            return True

        corners = obj.get_llur()
        x1, y1, x2, y2 = list(map(int, corners))
        old = self._get_state()
        self._update_crop_rect(x1, y1, x2, y2)
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="change crop"))
        self.pipeline.run_from(self)

    def _update_crop_rect(self, x1, y1, x2, y2):
        # reposition other elements to match
        if self.aspect is not None:
            x1, y1, x2, y2 = self._enforce_aspect(x1, y1, x2, y2)
        x1, y1, x2, y2 = list(map(int, [x1, y1, x2, y2]))
        rect = self.crop_obj.objects[0]
        rect.x1, rect.y1, rect.x2, rect.y2 = x1, y1, x2, y2
        text = self.crop_obj.objects[1]
        text.x, text.y = x1, y2 + 4
        self.viewer.redraw(whence=3)

        wd, ht = self._img_dims
        self.set_crop_rect(wd, ht, x1, y1, x2, y2)

    def set_crop_rect(self, wd, ht, x1, y1, x2, y2):
        x1p, y1p, x2p, y2p = [x1 / wd, y1 / ht, x2 / wd, y2 / ht]
        self.crop_rect = (x1p, y1p, x2p, y2p)

    def get_crop_rect_px(self, wd, ht, use_image_lim=False):
        x1p, y1p, x2p, y2p = np.array(self._crop_rect)
        x1, y1, x2, y2 = (int(x1p * wd), int(y1p * ht),
                          int(x2p * wd), int(y2p * ht))
        if use_image_lim:
            x1, y1, x2, y2 = max(0, x1), max(0, y1), min(x2, wd), min(y2, ht)
        return x1, y1, x2, y2

    def _set_aspect(self, asp_s):
        wd, ht = self._img_dims
        x1, y1, x2, y2 = self.get_crop_rect_px(wd, ht)

        if len(asp_s) == 0:
            self._aspect = None

        else:
            if ':' in asp_s:
                wd, ht = [float(n) for n in asp_s.split(':')]
                self._aspect = wd / ht
            else:
                self._aspect = float(asp_s)

        self._update_crop_rect(x1, y1, x2, y2)

    def _set_aspect_cb(self, widget):
        asp_s = widget.get_text().strip()
        old = self._get_state()
        self._set_aspect(asp_s)
        new = self._get_state()
        self.pipeline.push(StageAction(self, old, new,
                                       descr="set aspect ratio"))
        self.pipeline.run_from(self)

    def _enforce_aspect(self, x1, y1, x2, y2):

        ctr_x, ctr_y = (x1 + x2) / 2.0, (y1 + y2) / 2.0
        if self.aspect > 0:
            wd = x2 - x1
            hht = wd / self.aspect * 0.5
            ctr_y = (y1 + y2) * 0.5
            y1, y2 = ctr_y - hht, ctr_y + hht
        else:
            ht = y2 - y1
            hwd = ht / self.aspect * 0.5
            ctr_x = (x1 + x2) * 0.5
            x1, x2 = ctr_x - hwd, ctr_x + hwd

        return x1, y1, x2, y2

    def _get_state(self):
        return dict(crop_rect=[float(n) for n in self._crop_rect],
                    aspect=self._aspect)

    def run(self, prev_stage):
        data = self.pipeline.get_data(prev_stage)
        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        ht, wd = data.shape[:2]
        dims = (wd, ht)

        if self._img_dims != dims:
            self._img_dims = dims

        x1, y1, x2, y2 = self.get_crop_rect_px(wd, ht,
                                               use_image_lim=True)
        res_np = data[y1:y2, x1:x2, ...]

        if self.gui_up:
            _ht, _wd = res_np.shape[:2]
            ## try:
            ##     asp_s = trcalc.calc_aspect_str(_wd, _ht)
            ## except Exception as e:
            ##     # sometimes Numpy throws a NaN error here
            ##     asp_s = "{}:{}".format(_wd, _ht)
            asp = _wd / _ht
            s = "{}x{} ({})".format(_wd, _ht, asp)
            self.w.size.set_text(s)

        self.pipeline.send(res_np=res_np)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.crop_rect = d['crop_rect']
        self.aspect = d['aspect']



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from ginga import RGBMap, AutoCuts
from ginga.gw import Widgets

from .base import Stage, StageAction

count = 0


class Preview(Stage):

    _stagename = 'preview'

    def __init__(self):
        super().__init__()

        self.fv = None
        self.in_image = None
        self.image = None
        self._chname = ""

    def build_gui(self, container):
        self.fv = self.pipeline.get("fv")

        fr = Widgets.Frame("Preview")

        captions = [('Preview Channel:', 'label', 'Channel', 'entryset'),
                    ('Detach Image', 'button'),
                    ]
        w, b = Widgets.build_info(captions, orientation='vertical')
        self.w.update(b)

        b.channel.set_tooltip("Channel for preview image")
        b.channel.add_callback('activated', self.set_channel_cb)
        b.channel.set_text(self._chname)

        b.detach_image.set_tooltip("Detach the current image")
        b.detach_image.add_callback('activated', self.insert_image_cb)

        fr.set_widget(w)
        container.set_widget(fr)

    @property
    def chname(self):
        return self._chname

    @chname.setter
    def chname(self, val):
        self._chname = val
        if self.gui_up:
            self.w.channel.set_text(val)

    def get_channel(self, chname):
        channel = self.fv.get_channel_on_demand(chname)
        viewer = channel.viewer

        # PassThruRGBMapper does not do any RGB mapping
        rgbmap = RGBMap.PassThruRGBMapper(self.logger)
        viewer.set_rgbmap(rgbmap)

        # Clip cuts assumes data does not need to be scaled in cut levels--
        # only clipped
        viewer.set_autocuts(AutoCuts.Clip(logger=self.logger))

        return channel

    def set_channel_cb(self, widget):
        old = dict(chname=self._chname)
        self._chname = widget.get_text().strip()
        new = dict(chname=self._chname)
        self.pipeline.push(StageAction(self, old, new,
                                       descr="preview / change channel"))

        self.get_channel(self._chname)

        self.pipeline.run_from(self)

    def insert_image_cb(self, widget):
        image, self.image = self.image, None
        self.in_image = None

        self.pipeline.run_from(self)

    def run(self, prev_stage):
        global count
        data = self.pipeline.get_data(prev_stage)

        self.verify_2d(data)

        if self._bypass or data is None:
            self.pipeline.send(res_np=data)
            return

        if len(self.chname) > 0:
            self.pipeline.logger.info('pipeline preview')
            channel = self.fv.get_channel_on_demand(self._chname)

            in_image = self.pipeline.get('input_image')
            if in_image is not self.in_image:
                self.in_image = in_image
                # <-- new image.  Make one of the same type as the input
                # TODO: this needs to be user-selectable
                # TODO: if this is a revisited image, should look
                # up the corresponding previously generated output
                # image, if there is one and load it as the output.
                self.image = in_image.__class__(logger=self.pipeline.logger)

                # copy the header from the input
                in_header = in_image.get_header()
                # TODO: massage header, maybe add some metadata from
                # pipeline?
                self.image.update_keywords(in_header)

                # assign an output image name
                # TODO: name a better output name, that is some kind
                # of modified name of the input image name
                self.image.set(name='P' + str(count))
                count += 1
                self.image.set_data(data)

                channel.add_image(self.image)

            else:
                if self.image is not None:
                    self.image.set_data(data)

            # data has changed so redraw image completely
            channel.fitsimage.redraw(whence=0)

        self.pipeline.send(res_np=data)

    def get_image(self):
        return self.image

    def _get_state(self):
        return dict(chname=self._chname)

    def export_as_dict(self):
        d = super().export_as_dict()
        d.update(self._get_state())
        return d

    def import_from_dict(self, d):
        super().import_from_dict(d)
        self.chname = d['chname']
